"""Implements the grounding step of the KIE pipeline."""
import logging
from typing import List

import pandas as pd
from pyspark.sql.dataframe import DataFrame

from databricks.kie.data_utils import get_split_from_labeled, get_split_from_unlabeled
from databricks.kie.kie_state import KIEState
from databricks.kie.task_spec import KIETaskSpec
from databricks.model_serving.types.pt_endpoint import GenerationOutput, PayGoEndpoint
from databricks.model_serving.utils import MODEL_CONTEXT_LENGTH, ModelSpecs
from databricks.model_training.api.utils import get_spark

logger = logging.getLogger(__name__)


def calculate_fewshot_samples(tokens_processed: int, num_requests: int, current_samples: int) -> int:
    """
    Limit the number of fewshot examples based on the per-sample token average
    
    Context length is 131072 tokens, so we should leave plenty of room for the actual generation
    We'll leave 2 examples worth of space for the model to generate
    """
    tokens_per_example = tokens_processed / num_requests
    return max(min(current_samples, int(MODEL_CONTEXT_LENGTH / tokens_per_example) - 2), 0)


def create_grounding_df(
    base_df: pd.DataFrame,
    responses: List[GenerationOutput],
) -> DataFrame:
    spark = get_spark()
    response_strings, _ = zip(*responses)
    return spark.createDataFrame(pd.concat([base_df, pd.DataFrame(response_strings, columns=["response"])], axis=1))


def ground(task_spec: KIETaskSpec, state: KIEState, grounding_model: ModelSpecs) -> None:
    """
    Ground the task spec by running the grounding step.
    """
    if not state.requires_grounding:
        return

    if not grounding_model.endpoint:
        raise ValueError("Grounding model endpoint is not available")

    print('🌱 Selecting a few examples and extracting from them. This may take a moment...')
    if state.labeled_split_df:
        grounding_df = get_split_from_labeled(state.labeled_split_df, "grounding",
                                              task_spec.labeled_dataset_text_column,
                                              task_spec.labeled_dataset_output_json_column)
        logger.info(f"Got grounding df from labeled data. Using {grounding_df.count()} examples.")
    else:
        grounding_df = get_split_from_unlabeled(state.unlabeled_split_df, "grounding")
        logger.info(f"Got grounding df from unlabeled data. Using {grounding_df.count()} examples.")

    # Run inference using a strong model
    with PayGoEndpoint(grounding_model.endpoint) as endpoint:
        print("Extracting information from examples...")
        as_pd = grounding_df.toPandas()
        requests: List[str] = as_pd["request"].tolist()
        responses = endpoint.generate_batch(requests,
                                            system_prompt=state.ground_truth_prompt,
                                            response_format=state.response_format,
                                            show_progress=False)
        state.grounding_df = create_grounding_df(as_pd, responses)
        state.requires_grounding = False

        # Since we regenerated the grounding set, we'll need to update our downstream datasets
        state.requires_val = True
        state.requires_train_gen = True

        # Potentially limit fewshot examples based on size
        state.num_fewshot_samples = calculate_fewshot_samples(endpoint.tokens_processed, len(requests),
                                                              state.num_fewshot_samples)



"""Implements the optimization and data generation steps of the KIE pipeline"""
import json
import logging
import os
import time
from datetime import datetime
from functools import partial
from typing import Callable, ContextManager, Dict, List, Optional, Tuple, Type, cast

import mlflow
import pandas as pd
from mlflow.entities import Run as MlflowRun
from pydantic import BaseModel
from pyspark.sql.connect.dataframe import DataFrame

from databricks.kie.constants import MLFLOW_RESPONSE_FORMAT_FILE, MLFLOW_SYSTEM_PROMPT_FILE, MLFLOW_TRAILING_PROMPT_FILE
from databricks.kie.data_utils import (MIN_DOCS_TO_TRAIN, MIN_LABELED_TRAIN_SAMPLES, create_training_data,
                                       filter_valid_json, get_fewshot_examples, get_split_from_labeled,
                                       get_split_from_unlabeled, write_batch_to_cache)
from databricks.kie.inference_utils import format_ai_query
from databricks.kie.kie_evaluator import OVERALL_SCORE, evaluate_model
from databricks.kie.kie_schema import get_schema_hash
from databricks.kie.kie_state import KIEState
from databricks.kie.task_spec import KIETaskSpec
from databricks.model_serving.types.pt_endpoint import BaseEndpoint, PayGoEndpoint, ProvisionedThroughputEndpoint
from databricks.model_serving.utils import MODEL_CONTEXT_LENGTH, ModelSpecs, get_model_specs
from databricks.model_training.api import foundation_model as fm
from databricks.model_training.api.utils import get_schema_from_table, get_spark
from databricks.model_training.mlflow_utils import (add_notebook_source, change_run_name, get_mlflow_client,
                                                    log_delta_table_source)
from databricks.model_training.types.run_status import RunStatus
from databricks.model_training.types.training_run import TrainingRun

logger = logging.getLogger(__name__)

DOLLAR_LIMIT_ENV_VAR = "DOLLAR_LIMIT_TESTING"


def get_run_name(idx: int, schema_hash: str, suffix: str) -> str:
    return f"kie-optimized-{idx}{suffix}-{schema_hash}"


class BaseDataGenerator:
    """Base class for generating data for training and validation"""

    def __init__(
        self,
        task_spec: KIETaskSpec,
        state: KIEState,
        model: ModelSpecs,
        use_provisioned_throughput: bool = True,
    ):
        self.task_spec = task_spec
        self.state = state
        self.model = model
        self.use_provisioned_throughput = use_provisioned_throughput

    def _get_endpoint_context(self, teardown: bool = True) -> Callable[[], ContextManager[BaseEndpoint]]:
        """Get appropriate endpoint context based on configuration."""
        if self.use_provisioned_throughput:
            if not self.model.uc_schema:
                raise ValueError('UC schema required for provisioned throughput')
            return partial(ProvisionedThroughputEndpoint,
                           self.model.uc_schema,
                           model_version=self.model.version,
                           teardown_on_exit=teardown)

        if not self.model.endpoint:
            raise ValueError('Endpoint required for pay-as-you-go')
        return partial(PayGoEndpoint, self.model.endpoint)


class ValidationDataGenerator(BaseDataGenerator):
    """Generates validation data given the specified model, but only if there's no labeled data"""

    def __init__(
        self,
        task_spec: KIETaskSpec,
        state: KIEState,
        model: ModelSpecs,
        use_provisioned_throughput: bool = True,
        output_column: str = 'expected_response',
    ):
        super().__init__(task_spec, state, model, use_provisioned_throughput=use_provisioned_throughput)
        self.output_column = output_column

    def _get_labeled_validation_data(self) -> Optional[DataFrame]:
        """Get validation data from labeled dataset if available."""
        if not self.state.labeled_split_df:
            return None

        print("Using labeled data for validation set")
        return get_split_from_labeled(
            self.state.labeled_split_df,  # type: ignore
            'val',
            self.task_spec.labeled_dataset_text_column,
            self.task_spec.labeled_dataset_output_json_column)

    def _generate_responses(self, requests: List[str], endpoint: BaseEndpoint) -> List[str]:
        """Generate responses using the endpoint."""
        responses = endpoint.generate_batch(requests,
                                            system_prompt=self.state.ground_truth_prompt,
                                            response_format=self.state.response_format,
                                            show_progress=False)
        total_tokens = sum(r.total_tokens for r in responses)
        logger.info(f'Total tokens used during generation: {total_tokens}')
        return [r.response for r in responses]

    def generate(self) -> None:
        """Generate validation data from either labeled or unlabeled data."""
        if not self.state.requires_val:
            return

        print('Getting validation data...')
        # Try labeled data first
        val_df = self._get_labeled_validation_data()
        if val_df is None:
            print("Generating validation dataset...")
            val_df = get_split_from_unlabeled(self.state.unlabeled_split_df, "val")  # type: ignore
            print(f"Using {val_df.count()} documents from unlabeled dataset")

            # Generate responses
            endpoint_context = self._get_endpoint_context(teardown=False)
            if isinstance(endpoint_context, ProvisionedThroughputEndpoint):
                print("Creating an endpoint to hit. This may take a moment...")
            with endpoint_context() as endpoint:
                print("Extracting information from validation documents...")
                as_pd = val_df.toPandas()
                requests = as_pd["request"].tolist()
                responses = self._generate_responses(requests, endpoint)

                # Combine with original data
                combined_df = pd.concat([as_pd, pd.DataFrame(responses, columns=[self.output_column])], axis=1)
                spark = get_spark()
                val_df = spark.createDataFrame(combined_df)

        # Cache result
        self.state.val_df = filter_valid_json(val_df, self.output_column, self.state.response_format)
        self.state.requires_val = False


def eval_single_model(data_path: str,
                      model: ModelSpecs,
                      prompt: str,
                      experiment_id: str,
                      response_format: Type[BaseModel],
                      run_id: str,
                      trailing_prompt: Optional[str] = None) -> DataFrame:
    """
    Evaluate a single model using the provided validation data.

    Args:
        data_path: Path to validation data
        model: Model specifications
        prompt: Prompt to use for evaluation
        experiment_id: MLflow experiment ID
        response_format: Expected response format
        run_id: Optional MLflow run ID
        status: Optional status indicator for progress updates

    Returns:
        DataFrame containing evaluation results
    """

    if not model.uc_schema:
        raise ValueError('Model must have a UC schema defined')

    if trailing_prompt is None:
        trailing_prompt = prompt

    with ProvisionedThroughputEndpoint(model.uc_schema) as endpoint:
        ai_query = format_ai_query(data_path, endpoint, prompt, trailing_prompt=trailing_prompt)
        spark = get_spark()

        model_df = spark.createDataFrame(spark.sql(ai_query).toPandas())
        # Log eval dataset to run
        log_delta_table_source(run_id, data_path, name="Validation")

        model_df = evaluate_model(model_df, experiment_id, model, response_format, run_id=run_id)
        return model_df  # type: ignore - DataFrame vs connect DataFrame T_T


def get_mlflow_run_by_name(run_name: str, experiment_id: str) -> Optional[MlflowRun]:
    """
    Find MLflow run by name in the given experiment.

    Args:
        run_name: Name of the run to find
        experiment_id: MLflow experiment ID

    Returns:
        MlflowRun if found, None otherwise
    """
    runs: List[MlflowRun] = mlflow.search_runs(experiment_ids=[experiment_id],
                                               filter_string=f"run_name='{run_name}'",
                                               output_format='list')  # type: ignore
    return runs[0] if runs else None


def get_fewshot_target_column(df: DataFrame) -> str:
    """ Get the response column from the grounding data.
    Prefers "expected_response" over "response" if both are present.
    """
    for col in ['expected_response', 'response']:
        if col in df.columns:
            return col
    raise ValueError('Response column not found in grounding data')


def update_run_with_required_data(run_id: str, system_prompt: str, trailing_prompt: str,
                                  response_format: Type[BaseModel]) -> None:
    """ Log the required data on a run that we will use for our state
    
    We log:
    1. The system prompt
    2. The trailing prompt
    3. The response format
    
    We also tag the run with:
    1. The endpoint uc_schema  TODO: This is currently in the eval code but should be here
    2. The datasets
    """

    # Log the prompts
    mlflow.log_text(system_prompt, MLFLOW_SYSTEM_PROMPT_FILE, run_id=run_id)
    mlflow.log_text(trailing_prompt, MLFLOW_TRAILING_PROMPT_FILE, run_id=run_id)

    # Log the response format
    mlflow.log_text(json.dumps(response_format.model_json_schema(), indent=2),
                    MLFLOW_RESPONSE_FORMAT_FILE,
                    run_id=run_id)


def eval_baseline_models(task_spec: KIETaskSpec, state: KIEState, models: List[ModelSpecs]) -> None:
    """
    Evaluate baseline models in both zero-shot and few-shot settings.

    Args:
        task_spec: Task specification (unused)
        state: Current task state
        models: List of models to evaluate
    """
    del task_spec

    model_dfs = {}

    if not state.grounding_df:
        raise ValueError('Grounding data required for evaluation')

    response_column = get_fewshot_target_column(state.grounding_df)
    if state.num_fewshot_samples > 0:
        few_shot_examples = get_fewshot_examples(state.grounding_df,
                                                 state.num_fewshot_samples,
                                                 state.response_format,
                                                 response_column=response_column)
        fewshot_prompt = state.prompt_builder.build_prompt(few_shot_examples, include_markdown=True)
        run_fewshot = True
    else:
        fewshot_prompt = ""
        run_fewshot = False

    schema_hash = get_schema_hash(state.response_format)

    print("Evaluating a few initial models...")
    client = get_mlflow_client()
    for idx, model in enumerate(models, 1):
        # Zero-shot evaluation
        run_name = get_run_name(idx, schema_hash, 'A')
        existing_run = get_mlflow_run_by_name(run_name, state.experiment.experiment_id)

        if not existing_run or not run_has_eval(existing_run.info.run_id):
            run_id = existing_run.info.run_id if existing_run else None
            if not run_id:
                run = client.create_run(experiment_id=state.experiment.experiment_id, run_name=run_name)
                run_id = run.info.run_id
                add_notebook_source(run_id)
            update_run_with_required_data(run_id, state.zeroshot_prompt, state.zeroshot_prompt, state.response_format)

            print(f"Evaluating {run_name}...")
            df = eval_single_model(state.val_table_path, model, state.zeroshot_prompt, state.experiment.experiment_id,
                                   state.response_format, run_id)

            model_dfs[run_name] = df
        else:
            print(f"Skipping eval for run {run_name} because it already exists")

        # Few-shot evaluation
        if not run_fewshot:
            continue

        # Few-shot evaluation
        run_name = get_run_name(idx, schema_hash, 'B')
        existing_run = get_mlflow_run_by_name(run_name, state.experiment.experiment_id)

        if not existing_run or not run_has_eval(existing_run.info.run_id):
            run_id = existing_run.info.run_id if existing_run else None
            if not run_id:
                run = client.create_run(experiment_id=state.experiment.experiment_id, run_name=run_name)
                run_id = run.info.run_id
                add_notebook_source(run_id)
            update_run_with_required_data(run_id, fewshot_prompt, state.zeroshot_prompt, state.response_format)

            print(f"Evaluating {run_name}...")
            df = eval_single_model(state.val_table_path,
                                   model,
                                   fewshot_prompt,
                                   state.experiment.experiment_id,
                                   state.response_format,
                                   run_id,
                                   trailing_prompt=state.zeroshot_prompt)

            model_dfs[run_name] = df
        else:
            print(f"Skipping eval for run {run_name} because it already exists")

    if not state.model_dfs:
        state.model_dfs = {}
    state.model_dfs.update(**model_dfs)


class TrainingDataGenerator(BaseDataGenerator):
    """Generates training data given the specified model, but only if there's not enough labeled data
    """

    def __init__(
        self,
        task_spec: KIETaskSpec,
        state: KIEState,
        model: ModelSpecs,
        optimized_models: List[ModelSpecs],
        use_provisioned_throughput: bool = True,
        min_labeled_samples: int = MIN_LABELED_TRAIN_SAMPLES,
        # TODO: If we have different sweeps per model, we should just add the information to the model spec instead
        num_learning_rates_per_model: int = 3,  # {5e-7, 1e-6, 1e-5} so 3 learning rates in total
        total_epochs_per_model: int = 6,  # {2, 4} are the epoch sweep so 6 epochs in total
        dollar_cost: int = 500,
    ):
        super().__init__(task_spec, state, model, use_provisioned_throughput=use_provisioned_throughput)
        self.min_labeled_samples = min_labeled_samples
        self.optimized_models = optimized_models
        self.num_learning_rates_per_model = num_learning_rates_per_model
        self.total_epochs_per_model = total_epochs_per_model
        if DOLLAR_LIMIT_ENV_VAR in os.environ:
            logger.info(f"Using environment variable {DOLLAR_LIMIT_ENV_VAR} to "
                        f"set dollar cost: ${os.environ[DOLLAR_LIMIT_ENV_VAR]}")
            self.dollar_cost = int(os.environ[DOLLAR_LIMIT_ENV_VAR])
        else:
            self.dollar_cost = dollar_cost

    def _get_labeled_training_data(self) -> Optional[Tuple[List[str], List[str]]]:
        """Extract training data from labeled dataset if available and sufficient."""
        if not self.state.labeled_split_df:
            return None

        train_df = get_split_from_labeled(self.state.labeled_split_df, 'train',
                                          self.task_spec.labeled_dataset_text_column,
                                          self.task_spec.labeled_dataset_output_json_column)

        if train_df.count() < self.min_labeled_samples:
            print(f"Only {train_df.count()} labeled documents found. "
                  "Using unlabeled dataset instead.")
            return None

        print(f"Using {train_df.count()} labeled documents for training")
        return tuple(zip(*train_df.select("request", "response").toPandas().values.tolist()))

    def _get_token_budget(self) -> int:
        """Get the token budget for the training data generation.

        We are currently training each model using the same number of tokens, but this could be better if
        we use an equivalent cost budget for each model instead of a token budget.

        If we have
        - N (token budget in millions of tokens),
        - C1 (cost of model 1 per million tokens), C2 (cost of model 2 per million tokens), etc.
        - E (total epochs across all trials of a model)
        - D (total dollar cost)

        Then we have the following equation:

        N*(C1 + C2 + ... + CN)*E = D

        This gives us N in millions of tokens. Theoretically, we could round down to the nearest million, but
        if we hit a case where we round down to 0 million, we will not be able to train the model.
        """
        total_cost_per_m_ft_tokens = sum(model.cost_per_m_ft_tokens for model in self.optimized_models)
        tokens_in_millions = self.dollar_cost / (self.num_learning_rates_per_model * self.total_epochs_per_model *
                                                 total_cost_per_m_ft_tokens)
        return int(tokens_in_millions * 1_000_000)

    def generate(self) -> None:
        """Generate training data from either labeled or unlabeled data."""
        if not self.state.requires_train_gen:
            return

        generated = True
        print("Getting training data...")

        # Try labeled data first
        labeled_data = self._get_labeled_training_data()
        # Temporarily disable using labeled training data until we test it more and implement cost bounding
        if False and labeled_data:  # pylint: disable=condition-evals-to-constant
            requests, responses = labeled_data
            print("Preparing labeled data for training...")
            write_batch_to_cache(self.state.ground_truth_prompt, requests, responses, self.state.response_format,
                                 self.state.train_jsonl_path)
        else:
            # Fall back to unlabeled data
            print("Generating training data from unlabeled data")
            endpoint_context = self._get_endpoint_context()
            if isinstance(endpoint_context, ProvisionedThroughputEndpoint):
                print("Creating an endpoint to hit. This may take a moment...")

            with endpoint_context() as endpoint:
                print("Extracting information from training documents...")
                output = create_training_data(
                    unlabeled_split_df=self.state.unlabeled_split_df,
                    prompt=self.state.ground_truth_prompt,
                    response_format=self.state.response_format,
                    train_jsonl_cache_path=self.state.train_jsonl_path,
                    endpoint=endpoint,
                    token_budget=self._get_token_budget(),
                )
                if not output:
                    generated = False
                    print(f"Unlabeled dataset {self.task_spec.unlabeled_dataset} does not have enough " +
                          f"documents for training. We need at least {MIN_DOCS_TO_TRAIN + 25}")

        self.state.requires_train_gen = not generated


def launch_sweep(task_spec: KIETaskSpec, state: KIEState, models: List[ModelSpecs]) -> None:
    """Launch hyperparameter sweep.

    Launches training runs with different learning rates and epochs for each model.

    Args:
        task_spec: Task specification containing output location info
        state: Current task state with training data and experiment details
        models: List of models to sweep over

    Raises:
        ValueError: If models list is empty or no ft_model_name defined
        RuntimeError: If training data path or experiment name not set
    """

    if not state.train_jsonl_path:
        raise RuntimeError('Training data path not set')

    if not state.experiment.name:
        raise RuntimeError('Experiment name not set')

    register_to = '.'.join(task_spec.output_table.split('.')[:-1])
    suffixes = 'CDEFGHIJKLMNOPQRSTUVWXYZ'
    suffix_dict = dict(zip(range(1, len(suffixes) + 1), suffixes))
    schema_hash = get_schema_hash(state.response_format)
    for idx, model in enumerate(models, 1):
        if not model.ft_model_name:
            raise ValueError(f'Model {model} has no ft_model_name defined')

        sweep_idx = 0

        # TODO:(DK) Update the sweep params when we turn PEFT on
        for lr in [5e-7, 1e-6, 1e-5]:
            for ep in [2, 4]:
                sweep_idx += 1
                run_name = get_run_name(idx, schema_hash, suffix_dict[sweep_idx])
                existing_run = get_mlflow_run_by_name(run_name, state.experiment.experiment_id)
                if existing_run:
                    print(f"Skipping launching run {run_name} because it already exists")
                    continue

                run = fm.create(
                    task_type="CHAT_COMPLETION",
                    model=model.ft_model_name,
                    train_data_path=state.train_jsonl_path,
                    experiment_path=state.experiment.name,
                    register_to=f"{register_to}.{run_name}-{state.experiment.experiment_id}",
                    learning_rate=lr,
                    training_duration=f"{ep}ep",
                    context_length=MODEL_CONTEXT_LENGTH,
                )
                assert run.run_id is not None
                update_run_with_required_data(run.run_id, state.ground_truth_prompt, state.ground_truth_prompt,
                                              state.response_format)

                # Update the run name to obscure it
                change_run_name(run.run_id, run_name)

                # Add notebook source since it's not passed through
                add_notebook_source(run.run_id)


def get_active_runs(experiment_id: str) -> List[TrainingRun]:
    return [run for run in fm.list(experiment_id=experiment_id) if run.status.before(RunStatus.COMPLETED)]


def run_has_eval(run_id: str, schema_hash: Optional[str] = None) -> bool:
    run = mlflow.get_run(run_id=run_id)
    run_name = cast(str, run.info.run_name)
    if schema_hash and not run_name.endswith(schema_hash):
        # If this is not from the same schema hash, we can skip it
        logger.info(f"Skipping eval for run {run_name} because it is not from the current schema")
        return True
    return run.data.metrics.get(OVERALL_SCORE) is not None


def get_runs_without_eval(experiment_id: str, schema_hash: Optional[str] = None) -> List[TrainingRun]:
    completed_runs = fm.list(experiment_id=experiment_id, statuses=[RunStatus.COMPLETED])
    return [run for run in completed_runs if run.run_id and not run_has_eval(run.run_id, schema_hash)]


class SweepEvaluator:
    """Follows runs in a sweep and evaluates them when they complete
    """

    def __init__(self, task_spec: KIETaskSpec, state: KIEState, timeout_minutes: int = 720):
        self.task_spec = task_spec
        self.state = state
        self.timeout_minutes = timeout_minutes
        self.start_time = datetime.now()

    def _get_model_mapping(self) -> Dict[str, ModelSpecs]:
        model_specs = get_model_specs()
        return {model.ft_model_name: model for model in model_specs.values() if model.ft_model_name is not None}

    def _create_model_details(self, run: MlflowRun, base_model: ModelSpecs) -> ModelSpecs:
        output_schema = get_schema_from_table(self.task_spec.output_table)
        run_name = cast(str, run.info.run_name)
        return ModelSpecs(
            name=run_name,
            endpoint='',
            cost_per_hour=base_model.cost_per_hour,
            ft_model_name=base_model.ft_model_name,
            is_hosted=False,
            uc_schema=f'{output_schema}.{run_name}-{self.state.experiment.experiment_id}',
        )

    def _eval_run(self, run: MlflowRun, model: ModelSpecs) -> DataFrame:
        model_details = self._create_model_details(run, model)

        return eval_single_model(self.state.val_table_path, model_details, self.state.ground_truth_prompt,
                                 self.state.experiment.experiment_id, self.state.response_format, run.info.run_id)

    def _is_timed_out(self) -> bool:
        elapsed = datetime.now() - self.start_time
        return elapsed.total_seconds() > (self.timeout_minutes * 60)

    def _check_run_deleted(self, run_id: str) -> bool:
        run = mlflow.get_run(run_id=run_id)
        if run.info.lifecycle_stage == "deleted":
            return True
        return False

    def evaluate_sweep(self) -> None:
        remaining = get_active_runs(self.state.experiment.experiment_id)
        schema_hash = get_schema_hash(self.state.response_format)
        run_eval = [
            r for r in get_runs_without_eval(self.state.experiment.experiment_id, schema_hash)
            if not self._check_run_deleted(r.run_id)
        ]  # type: ignore

        if not remaining and not run_eval:
            if not self.state.model_dfs:
                self.state.model_dfs = {}
            return

        model_map = self._get_model_mapping()
        run_to_model = {run.name: model_map.get(run.model or '') for run in remaining + run_eval}

        model_dfs = {}

        n_models = len(remaining)
        last_update = len(remaining)
        while not self._is_timed_out():
            logger.info(f"Checking runs: {remaining}")
            if len(remaining) < last_update:
                print(f"({n_models - len(remaining)}/{n_models}) Waiting for models to train...")
                last_update = len(remaining)
            # Process completed runs
            still_running = []
            for run in remaining:
                run = fm.get(run)
                if run.status is RunStatus.COMPLETED:
                    print(f"✓ Run {run.name} completed. Beginning evaluation")
                    run_eval.append(run)
                elif run.status is RunStatus.FAILED:
                    print(f"⨯ Run {run.name} failed")
                elif run.status is RunStatus.STOPPED:
                    print(f"⚠ Run {run.name} was canceled")
                else:
                    still_running.append(run)

            remaining = still_running

            # Evaluate completed runs
            if run_eval:
                print(f"Running evaluations for {len(run_eval)} models")
                if len(run_eval) > 1:
                    print("Evaluations are run in sequence, so this may take a moment")

            for run in run_eval:
                model = run_to_model[run.name]
                if model is None:
                    continue

                if not run.run_id:
                    raise RuntimeError(f"Run {run.name} has no run ID")

                if self._check_run_deleted(run.run_id):
                    # The run has been deleted
                    logger.info(f"Skipping eval for run {run.name} because it has been deleted")
                    continue

                mlflow_run = mlflow.get_run(run.run_id)
                print(f"Evaluating {mlflow_run.info.run_name}")
                model_dfs[mlflow_run.info.run_name] = self._eval_run(mlflow_run, model)
                print(f"Evaluations completed for {run.name}")  # TODO (TL): Add link to run evaluations page

            run_eval = []
            if not remaining:
                break

            time.sleep(60)

        self.state.model_dfs = model_dfs



"""Container for constants used in the KIE package"""
MLFLOW_REGISTERED_MODEL_TAG = "registered_to"
MLFLOW_SYSTEM_PROMPT_FILE = 'system_prompt.txt'
MLFLOW_TRAILING_PROMPT_FILE = 'trailing_prompt.txt'
MLFLOW_RESPONSE_FORMAT_FILE = 'response_format.json'



"""Inference utility functions to invoke LLM's using ChatCompletionsRequest."""
import json
from typing import Any, Dict, List, Optional

import requests as request_lib
from pydantic import BaseModel

from databricks.model_serving.types.pt_endpoint import BaseEndpoint
from databricks.model_training.api.utils import get_host_and_token_from_env, normalize_table_name

try:
    from openai.lib._parsing._completions import type_to_response_format_param
    OPENAI_AVAILABLE = True
except ImportError:
    OPENAI_AVAILABLE = False

# TODO(jun.choi): Make client ID configurable per usecase.
CLIENT_ID = 'finetuning-t2t'

# Constants
MARKDOWN_JSON_PATTERN = r'(?s).*?```(?:json)?\s*(.*?)\s*```\s*.*'
DOCUMENT_TEMPLATE = "\\n\\nDocument:\\n\\n"


def clean_request_text(text_col: str) -> str:
    """Create SQL to clean request text.
    
    Args:
        text_col: Column containing text to clean
    """
    return rf"""REPLACE(REPLACE(REPLACE({text_col}, '"', ''), '“', ''), '”', '')"""


def format_ai_query(
    source: str,
    endpoint: BaseEndpoint,
    system_prompt: str,
    max_tokens: int = 1024,
    temperature: float = 0.3,
    trailing_prompt: Optional[str] = None,
) -> str:
    """Format SQL query for AI extraction.
    
    Args:
        source: Source table name
        endpoint: Endpoint for AI queries
        system_prompt: System prompt for extraction

        
    Example:
        >>> query = format_ai_query(
        ...     source="my_table",
        ...     endpoint=my_endpoint,
        ...     system_prompt="Extract entities",
        ...     max_tokens=2048,
        ...     temperature=0.5,
        ... )
    
    Returns:
        Formatted SQL query string
    """

    # Handle table name quoting
    table_name = normalize_table_name(source)

    # Handle trailing prompt
    trailing_prompt = trailing_prompt or system_prompt

    # Build query components
    model_params = f"named_struct('max_tokens', {max_tokens}, 'temperature', {temperature})"
    clean_request = clean_request_text("request")

    full_prompt = f"""
        system_prompt || 
        '{DOCUMENT_TEMPLATE}' ||
        {clean_request} ||
        '\\n\\n' || 
        trailing_prompt
    """

    return f"""
    SELECT 
        *,
        CASE 
            WHEN extracted_response != '' THEN extracted_response 
            ELSE raw_response 
        END AS response
    FROM (
        SELECT 
            *,
            ai_query(
                '{endpoint.endpoint.name}',
                {full_prompt},
                failOnError => true,
                modelParameters => {model_params}
            ) AS raw_response,
            regexp_extract(raw_response, r'{MARKDOWN_JSON_PATTERN}', 1) AS extracted_response
        FROM {table_name}
        CROSS JOIN (
            SELECT 
                '{system_prompt}' AS system_prompt,
                '{trailing_prompt}' AS trailing_prompt
        )
    )
    """


def _create_request_db(chat_completions, model_id, timeout=None):
    api_url, api_token = get_host_and_token_from_env()
    headers = {
        'Authorization': f'Bearer {api_token}',
        'Content-Type': 'application/json',
    }
    api_path = f'{api_url}/api/2.0/conversation/proxy/chat/completions'
    payload = {
        '@method': 'databricksChatCompletionRequest',
        'params': {
            **chat_completions
        },
        'model': model_id,
        'metadata': {
            'clientId': CLIENT_ID
        },
    }
    return request_lib.post(api_path, headers=headers, json=payload, timeout=timeout)


def _create_request_ext(chat_completions, model_id, timeout=None):
    api_url, api_token = get_host_and_token_from_env()
    headers = {
        'Authorization': f'Bearer {api_token}',
        'Content-Type': 'application/json',
    }
    api_path = f'{api_url}/api/2.0/conversation/proxy/chat/completions'

    payload = {
        '@method': 'openAiServiceChatCompletionRequest',
        'params': {
            **chat_completions
        },
        'model': model_id,
        'apiVersion': '2024-09-01-preview',
        'metadata': {
            'clientId': CLIENT_ID
        },
    }
    return request_lib.post(api_path, headers=headers, json=payload, timeout=timeout)


_MODEL_FN_DICT = {
    # Common / shared models
    'databricks-meta-llama-3-70b-instruct': _create_request_db,
    'llama-3-1-405b': _create_request_db,
    'gpt-4o-2024-08-06': _create_request_ext,
    'gpt-4o-mini-2024-07-18': _create_request_ext,
    'gpt-4-0613': _create_request_ext,
    'gpt-4-turbo-2024-04-09': _create_request_ext,
    'gpt-35-turbo-1106': _create_request_ext,

    # Client specific models
    'gpt-4o-2024-08-06-text2text': _create_request_ext,
    'gpt-4o-mini-2024-07-18-text2text': _create_request_ext,
}


def get_llm_proxy_chat_completion_response(model_id: str, req: Dict, timeout: Optional[float] = None) -> str:
    """Retrieves the chat completion response from a specified language model.
    
        Args:
            model_id (str): The identifier of the language model to use.
            req (dict): The request payload to be sent to the language model.
            timeout (float, optional): The timeout for the request. Defaults to None.

        Returns:
            str: The content of the chat completion response.
    """
    if model_id not in _MODEL_FN_DICT:
        raise ValueError(f'Model id {model_id} is not supported. Please use one of {_MODEL_FN_DICT.keys()}')

    llm_proxy_model_id = model_id
    response = _MODEL_FN_DICT[model_id](req, llm_proxy_model_id, timeout=timeout)

    if response.status_code != 200:
        raise RuntimeError(f'Request failed with status code {response.status_code}: {response.text}')

    out = json.loads(response.json()['completion'])['choices'][0]['message']['content']
    return out


def generate_base_model_using_chat_completion_messages(messages: List[Dict[str, str]],
                                                       model_clz: Any,
                                                       model_id: Optional[str] = None) -> BaseModel:
    """Generates a base model using chat completion messages using structured output.

    Args:
        messages (list): A list of dictionaries representing the chat messages.
        model_clz (Any): The Pydantic model class to use for the response.
        model_id_suffix (str, optional): The suffix to append to the model id. Defaults to None.

    Returns:
        BaseModel: The generated Pydantic model.
    """
    if not OPENAI_AVAILABLE:
        raise ImportError("openai library is not available. Please install it to use this")

    req: Dict = {
        'messages': messages,
        'response_format': type_to_response_format_param(model_clz),
    }
    llm_proxy_model_id = model_id or 'gpt-4o-2024-08-06'
    res = get_llm_proxy_chat_completion_response(llm_proxy_model_id, req)
    return model_clz(**json.loads(res))



"""Module representing the schema / dataclass for Text2Text module."""
import itertools
from dataclasses import dataclass, field
from enum import Enum
from typing import Dict, List, Optional, Tuple

import mlflow
from pydantic import BaseModel, Field

from databricks.kie.data_utils import read_from_jsonl, read_from_table_with_columns

MAX_EXAMPLES_PER_DATA_SOURCE = 200


@dataclass
class JsonlFileDataConfig:
    """Jsonl file data configuration"""
    file_path: str
    input_key: str
    output_key: str


@dataclass
class UCTableDataConfig:
    """Unity catalog table data configuration"""
    table_path: str
    input_column_name: str
    output_column_name: str


@dataclass
class Text2TextTaskSpec:
    """Task spec for Text2Text experiments"""
    instruction: str
    experiment_name: Optional[str] = None
    json_examples: Optional[List[Dict[str, str]]] = field(default_factory=list)
    uc_table: Optional[UCTableDataConfig] = None
    jsonl_file: Optional[JsonlFileDataConfig] = None
    evaluation_criteria: Optional[List[str]] = None

    def __post_init__(self):
        if not self.json_examples and not self.uc_table and not self.jsonl_file:
            raise ValueError("Missing data config for Text2Text -- one of `json_examples`, `uc_table`, `jsonl_file`"
                             " should be provided")

        return self

    def to_system_param(self) -> 'T2TSystemParams':
        """Converts the task spec to system params for a Text2Text model"""
        # TODO(jun.choi): Support unlabelled examples.
        labelled_examples = []
        if self.uc_table:
            columns = (self.uc_table.input_column_name, self.uc_table.output_column_name)
            table_data = read_from_table_with_columns(self.uc_table.table_path, columns, MAX_EXAMPLES_PER_DATA_SOURCE)
            for datum in table_data:
                labelled_examples.append(
                    (datum[self.uc_table.input_column_name], datum[self.uc_table.output_column_name]))

        if self.json_examples:
            for datum in self.json_examples:
                if not set(datum.keys()) >= {'request', 'response'}:
                    raise ValueError("Invalid json example format. Should have keys 'request' and 'response'")

                labelled_examples.append((datum['request'], datum['response']))

        if self.jsonl_file:
            data = read_from_jsonl(self.jsonl_file.file_path, MAX_EXAMPLES_PER_DATA_SOURCE)
            for datum in data:
                labelled_examples.append((datum[self.jsonl_file.input_key], datum[self.jsonl_file.output_key]))

        return T2TSystemParams(instruction=self.instruction,
                               labelled_examples=labelled_examples,
                               evaluation_criteria=self.evaluation_criteria)


@dataclass
class T2TSystemParams:
    """System parameters for a T2T model."""
    instruction: str
    unlabelled_examples: Optional[List[str]] = None
    labelled_examples: Optional[List[Tuple[str, str]]] = None
    evaluation_criteria: Optional[List[str]] = None

    def __post_init__(self):
        if not self.unlabelled_examples and not self.labelled_examples:
            raise ValueError("One of labelled or unlabelled examples should be provided")
        return self

    def get_sxs_data(self, other: 'T2TSystemParams') -> List[Dict[str, str]]:
        """Get side by side data that can be used to comapre this and another T2TSystemParams."""
        common_unlabelled = set(self.unlabelled_examples or []).intersection(set(other.unlabelled_examples or []))
        common_labelled = set(self.labelled_examples or []).intersection(set(other.labelled_examples or []))
        data = []
        for datum in sorted(itertools.chain(common_unlabelled, common_labelled)):
            if isinstance(datum, str):
                data.append({'request': datum})
            else:
                data.append({'request': datum[0], 'response': datum[1]})
        return data

    def get_diff(self, other: 'T2TSystemParams') -> 'T2TSystemParamsDiff':
        """Get the differences between this and another T2TSystemParams.

        Returns: T2TSystemParamsDiff object containing fields (self - other).
        """
        diff = {}
        if self.instruction != other.instruction:
            diff['instruction'] = self.instruction
            diff['previous_instruction'] = other.instruction

        evaluation_criteria_diff = list(set(self.evaluation_criteria or []) - set(other.evaluation_criteria or []))
        if evaluation_criteria_diff:
            diff['evaluation_criteria'] = evaluation_criteria_diff

        added_labeled_example_diff = (list(set(self.labelled_examples or []) - set(other.labelled_examples or [])))
        removed_labeled_example_diff = (list(set(other.labelled_examples or []) - set(self.labelled_examples or [])))
        added_unlabeled_example_diff = (
            list(set(self.unlabelled_examples or []) - set(other.unlabelled_examples or [])))
        removed_unlabeled_example_diff = (
            list(set(other.unlabelled_examples or []) - set(self.unlabelled_examples or [])))
        diff['added_labelled_examples'] = added_labeled_example_diff
        diff['removed_labelled_examples'] = removed_labeled_example_diff
        diff['added_unlabelled_examples'] = added_unlabeled_example_diff
        diff['removed_unlabelled_examples'] = removed_unlabeled_example_diff
        return T2TSystemParamsDiff(**diff)


@dataclass
class T2TSystemParamsDiff:
    """System parameters diff between 2 versions of T2T model."""
    previous_instruction: Optional[str] = None
    instruction: Optional[str] = None
    evaluation_criteria: Optional[List[str]] = None
    added_labelled_examples: Optional[List[Tuple[str, str]]] = None
    removed_labelled_examples: Optional[List[Tuple[str, str]]] = None
    added_unlabelled_examples: Optional[List[str]] = None
    removed_unlabelled_examples: Optional[List[str]] = None


@dataclass
class SideBySideResult:
    """Side by side results for a T2T model"""
    request: str
    previous_response: str
    current_response: str


class InstructionType(str, Enum):
    classification = "classification"
    grounded_generation = "grounded_generation"
    subspan_extraction = "subspan_extraction"
    summarization = "summarization"
    others = "others"


# TODO(jun.choi): Add `optimized_instruction` in InstructionSchema via DSPY
class InstructionInfo(BaseModel):
    optimized_instruction: str = Field(
        description="""Instruction optimized for the model. Includes detailed guidelines including
        1. Instruction / task description.
        2. Nature of the input to be provided.
        3. Expected response format.
        4. Constraints on the response or criteria tat the response should satisfy.
        """)
    instruction_type: InstructionType = Field(description=("Task type of of the instruction. Can be one of :"
                                                           f"{ ', '.join(str(v.name) for v in InstructionType) }"))
    output_format: str = Field(
        description=("The format of the output response. For example "
                     "'Output should be a comma separated list' / 'Response should be as concise as possible' "))


@dataclass
class T2TSystemResult:
    """Model evaluation result for a T2T model"""
    eval_result: mlflow.models.EvaluationResult
    examples: List[Tuple[str, str]]
    sxs_result: Optional[List[SideBySideResult]]
    current_task_spec: Text2TextTaskSpec
    experiment_id: str
    model_candidates: List[str]
    default_model: str
    instruction_info: InstructionInfo



"""Evaluator class for Key Information Extraction (KIE) tasks."""

import json
import logging
import math
import re
import statistics
from typing import Any, Optional, Type, Union

import mlflow
import numpy as np
import pandas as pd
from databricks.agents.evals import metric  # pylint: disable=ungrouped-imports
from mlflow.evaluation import Assessment
from mlflow.models import EvaluationResult
from pydantic import BaseModel, ValidationError
from pydantic.fields import FieldInfo
from pyspark.sql.dataframe import DataFrame
from scipy.optimize import linear_sum_assignment

from databricks.kie.constants import MLFLOW_REGISTERED_MODEL_TAG
from databricks.kie.eval_utils import PrimitiveTypes, correctness, exact_match, fuzzy_match, normalized_match
from databricks.model_serving.utils import ModelSpecs
from databricks.model_training.api.utils import get_spark
from databricks.model_training.logging import temp_log_level

OVERALL_SCORE = 'overall_score'
OVERALL_SCORE_RATIONALE = 'Overall score is the average of the individual field scores'
IS_SCHEMA_MATCH_RATIONALE = 'The schema of the response matches the expected schema.'
IS_JSON_PARSABLE_RATIONALE = 'The response is valid JSON.'
IS_NOT_SCHEMA_MATCH_RATIONALE = 'The schema of the response does not match the expected schema.'
IS_NOT_JSON_PARSABLE_RATIONALE = 'The response is not valid JSON.'

# TODO figure out how enums come in
AllowedInnerTypes = Union[PrimitiveTypes, BaseModel, list['AllowedInnerTypes']]
FieldTuple = tuple[AllowedInnerTypes, FieldInfo]

# TODO: Replace correctness with guideline adherence

KIE_EVAL_METRIC_REGISTRY = {
    'exact_match': exact_match,
    'normalized_match': normalized_match,
    'fuzzy_match': fuzzy_match,
    'correctness': correctness,
}
DISALLOWED_NESTED_METRICS = {'correctness'}


def get_default_metric_from_field(field_info: FieldInfo, gt_value: PrimitiveTypes, from_root: bool) -> str:
    """Select a default metric to use based on the field information

    Note: In the future, this routing could be done by an LLM, or other more sophisticated method.

    Args:
        field_info (FieldInfo): Pydantic field information.
        gt_value (PrimitiveTypes): Ground truth value.
        from_root (bool): Whether the field is a root field.

    Returns:
        str: Default metric to use.
    """

    del field_info  # Currently unused, but might be used for the description in the future

    if not isinstance(gt_value, str):
        return 'exact_match'

    gt_length = len(gt_value)

    # Completely arbitrary threshold
    if gt_length <= 10:
        return 'normalized_match'
    elif not from_root or gt_length <= 75:
        return 'fuzzy_match'
    else:
        return 'correctness'


def parse_json_markdown(text: str) -> str:
    """Parse JSON from a markdown-formatted string.

    Allows text before or after the JSON block, and if no
    JSON block is found, returns the entire text.

    Args:
        text (str): Markdown-formatted string.

    Returns:
        str: JSON string.
    """
    pattern = r'^.*?```(?:json)?\s*(.*?)```\s*.*$'
    match = re.search(pattern, text, re.DOTALL)
    if match:
        return match.group(1).strip()
    return text.strip()


def is_null_like(value: Any) -> bool:
    """Check if a value is null-like.

    Checks None, empty dict, empty list, and various string representations of null

    Args:
        value (any): Value to check.

    Returns:
        bool: True if value is null-like, False otherwise.
    """
    null_values = [None, '', 'null', 'none', 'nan', 'n/a', 'na', [], {}]
    is_none = value is None
    is_null_value = value in null_values
    lower_is_null_value = isinstance(value, str) and value.lower() in null_values
    is_nan = isinstance(value, float) and math.isnan(value)
    return is_none or is_null_value or lower_is_null_value or is_nan


class KIEEvaluator:
    """Evaluator for text to json output."""

    def __init__(self, schema: Type[BaseModel]):
        self.schema = schema

    def evaluate_row(self, query: str, ground_truth_answer: str,
                     generated_answer: str) -> tuple[dict[str, Any], dict[str, Any]]:
        top_level_keys = self.schema.model_fields.keys()

        # Parse the provided strings into JSON strings
        gt_json_parsed = parse_json_markdown(ground_truth_answer)
        pred_json_parsed = parse_json_markdown(generated_answer)

        # Ground truth must be valid JSON
        try:
            gt_json_loaded = json.loads(gt_json_parsed)
        except json.JSONDecodeError as e:
            raise ValueError(f'Ground truth answer is not a valid JSON: {ground_truth_answer}') from e

        # Ground truth must match the provided schema
        try:
            gt_loaded = self.schema(**gt_json_loaded)
        except ValidationError as e:
            raise ValueError(f'Ground truth answer does not match the provided schema: {ground_truth_answer}') from e

        # Prediction may not have been valid JSON. Return early if it is not.
        try:
            pred_json_loaded = json.loads(pred_json_parsed)
        except json.JSONDecodeError as e:
            return {
                k: Assessment(name=k, value=0.0, rationale=IS_NOT_JSON_PARSABLE_RATIONALE) for k in top_level_keys
            }, {
                OVERALL_SCORE:
                    Assessment(name=OVERALL_SCORE, value=0.0, rationale=IS_NOT_JSON_PARSABLE_RATIONALE),
                'is_schema_match':
                    Assessment(name='is_schema_match', value=False, rationale=IS_NOT_JSON_PARSABLE_RATIONALE),
                'is_json_parsable':
                    Assessment(name='is_json_parsable', value=False, rationale=IS_NOT_JSON_PARSABLE_RATIONALE)
            }

        # Prediction may not match the provided schema. Return early if it is not.
        try:
            pred_loaded = self.schema(**pred_json_loaded)
        except ValidationError as e:
            return {
                k: Assessment(name=k, value=0.0, rationale=IS_NOT_SCHEMA_MATCH_RATIONALE) for k in top_level_keys
            }, {
                OVERALL_SCORE:
                    Assessment(name=OVERALL_SCORE, value=0.0, rationale=IS_NOT_SCHEMA_MATCH_RATIONALE),
                'is_schema_match':
                    Assessment(name='is_schema_match', value=False, rationale=IS_NOT_SCHEMA_MATCH_RATIONALE),
                'is_json_parsable':
                    Assessment(name='is_json_parsable', value=True, rationale=IS_NOT_SCHEMA_MATCH_RATIONALE)
            }

        # Both ground truth and prediction have been loaded into the provided schema and are pydantic objects.
        # TODO: Add an assert that the structures match at this point since all fields are required

        scored_output = self._scoring_helper_object(query, gt_loaded, pred_loaded, is_root=True)

        overall_score = statistics.mean([float(v.value) for v in scored_output.values()])
        overall_scoring = {
            OVERALL_SCORE: Assessment(name=OVERALL_SCORE, value=overall_score, rationale=OVERALL_SCORE),
            'is_schema_match': Assessment(name='is_schema_match', value=True, rationale=IS_SCHEMA_MATCH_RATIONALE),
            'is_json_parsable': Assessment(name='is_json_parsable', value=True, rationale=IS_JSON_PARSABLE_RATIONALE)
        }
        return scored_output, overall_scoring

    def _compute_pairwise_scores(
        self,
        query: str,
        gt_items: list[AllowedInnerTypes],
        pred_items: list[AllowedInnerTypes],
        gt_field_info: FieldInfo,
    ) -> np.ndarray:
        """Compute pairwise similarity scores between ground truth and predicted items."""
        scores = np.zeros((len(gt_items), len(pred_items)))

        for i, gt_item in enumerate(gt_items):
            for j, pred_item in enumerate(pred_items):
                scores[i, j] = self._scoring_helper_recurse(query, (gt_item, gt_field_info),
                                                            (pred_item, gt_field_info)).value

        return scores

    def _compute_assignment(self, scores: np.ndarray) -> tuple[list[int], list[int]]:
        """Find optimal assignment using Hungarian algorithm."""
        row_ind, col_ind = linear_sum_assignment(scores, maximize=True)
        return row_ind.tolist(), col_ind.tolist()

    def _compute_precision_recall_f1(self, scores: np.ndarray,
                                     assignments: tuple[list[int], list[int]]) -> tuple[float, float, float]:
        """Compute precision and recall based on assignments."""
        row_ind, col_ind = assignments
        sum_scores = scores[row_ind, col_ind].sum()

        num_gt_items = scores.shape[0]
        num_pred_items = scores.shape[1]

        precision = sum_scores / num_pred_items
        recall = sum_scores / num_gt_items
        f1 = 2 * precision * recall / (precision + recall) if (precision + recall) > 0 else 0

        return precision, recall, f1

    def _scoring_helper_object(self,
                               query: str,
                               ground_truth_answer: BaseModel,
                               generated_answer: BaseModel,
                               is_root: bool = False) -> dict[str, Assessment]:
        results = {}
        for key_name in ground_truth_answer.model_dump():
            gt_value = getattr(ground_truth_answer, key_name)
            pred_value = getattr(generated_answer, key_name)

            gt_field_info = ground_truth_answer.model_fields[key_name]
            pred_field_info = generated_answer.model_fields[key_name]

            score = self._scoring_helper_recurse(query, (gt_value, gt_field_info), (pred_value, pred_field_info),
                                                 from_root=is_root)
            results[key_name] = Assessment(
                name=f'{key_name}',
                value=score.value,
                rationale=score.rationale,
            )

        return results

    def _scoring_helper_recurse(self,
                                query: str,
                                ground_truth_answer: FieldTuple,
                                generated_answer: FieldTuple,
                                from_root: bool = False) -> Assessment:
        gt_value, gt_field_info = ground_truth_answer

        pred_value = generated_answer[0]

        # Early exit if either value is null-like
        gt_is_null_like = is_null_like(gt_value)
        pred_is_null_like = is_null_like(pred_value)
        if gt_is_null_like and pred_is_null_like:
            return Assessment(name='null_match', value=True, rationale='Both values are null-like')
        elif gt_is_null_like or pred_is_null_like:
            is_null_name = 'response' if pred_is_null_like else 'expected_response'
            is_not_null_name = 'expected_response' if pred_is_null_like else 'response'
            return Assessment(name='null_match',
                              value=False,
                              rationale=f'{is_null_name} is null-like but {is_not_null_name} is not')

        if isinstance(gt_value, PrimitiveTypes):
            assert isinstance(pred_value, PrimitiveTypes)

            default_kie_eval_metric = get_default_metric_from_field(gt_field_info, gt_value, from_root)

            if gt_field_info.json_schema_extra is not None:
                eval_metric_string = str(
                    gt_field_info.json_schema_extra.get(  # type: ignore
                        'eval_metric', default_kie_eval_metric))
                eval_metric_kwargs = gt_field_info.json_schema_extra.get('eval_metric_kwargs', {})  # type: ignore
            else:
                eval_metric_string = default_kie_eval_metric
                eval_metric_kwargs = {}

            if eval_metric_string not in KIE_EVAL_METRIC_REGISTRY:
                raise ValueError(f'Unsupported evaluation metric: {eval_metric_string}')

            eval_metric_func = KIE_EVAL_METRIC_REGISTRY[eval_metric_string]

            assert isinstance(eval_metric_kwargs, dict)

            if eval_metric_string in DISALLOWED_NESTED_METRICS and not from_root:
                raise ValueError(f'Unsupported evaluation metric {eval_metric_string} for nested field')

            score = eval_metric_func(query, gt_value, pred_value, **eval_metric_kwargs)
            return score
        elif isinstance(gt_value, BaseModel):
            assert isinstance(pred_value, BaseModel)
            per_key_results = self._scoring_helper_object(query, gt_value, pred_value)
            result = statistics.mean([float(v.value) for v in per_key_results.values()])
            return Assessment(name='object_score', value=result, rationale='Average of the individual field scores')
        elif isinstance(gt_value, list):
            assert isinstance(pred_value, list)
            pairwise_scores = self._compute_pairwise_scores(query, gt_value, pred_value, gt_field_info)
            assignments = self._compute_assignment(pairwise_scores)
            _, _, f1 = self._compute_precision_recall_f1(pairwise_scores, assignments)
            return Assessment(name='list_score',
                              value=f1,
                              rationale='Soft F1 score of the optimal pairwise assignment between list items')
        else:
            raise ValueError(f'Unsupported type: {type(gt_value)}')


def store_evaluation_metrics(df: pd.DataFrame, response_format: Type[BaseModel]) -> EvaluationResult:
    eval_class = KIEEvaluator(response_format)

    @metric
    def kie(request, response, expected_response):
        eval_output_per_key, eval_output_overall = eval_class.evaluate_row(request, expected_response, response)
        all_assessments = [eval_output_overall['is_schema_match']] + list(
            eval_output_per_key.values()) + [eval_output_overall[OVERALL_SCORE]]
        return all_assessments

    eval_result = mlflow.evaluate(
        model=None,
        data=df,
        model_type='databricks-agent',
        evaluator_config={"databricks-agent": {
            "metrics": [],
        }},
        extra_metrics=[kie],
    )

    return eval_result


def evaluate_model(df: DataFrame,
                   experiment_id: str,
                   model: ModelSpecs,
                   response_format: type[BaseModel],
                   run_name: Optional[str] = None,
                   run_id: Optional[str] = None) -> DataFrame:

    for required_column in ("request", "response", "expected_response"):
        if required_column not in df.columns:
            raise ValueError(f"`{required_column}` column is required for evaluation")

    model_df = df.toPandas()

    # Log each metric to mlflow
    start_kwargs = {}
    if run_id:
        start_kwargs["run_id"] = run_id
    elif run_name:
        start_kwargs["run_name"] = run_name
    with mlflow.start_run(experiment_id=experiment_id, **start_kwargs), temp_log_level("mlflow", logging.ERROR):
        mlflow.log_param(key="cost", value=model.cost_per_hour)
        mlflow.set_tag(MLFLOW_REGISTERED_MODEL_TAG, model.uc_schema)
        eval_result = store_evaluation_metrics(model_df, response_format)
        result_df = eval_result.tables['eval_results']

    spark = get_spark()
    return spark.createDataFrame(result_df)



"""Module implementaion of entrypoint API for Text2Text."""

import functools
import random
import re
from datetime import datetime
from typing import Dict, List, Optional

import mlflow
import pandas as pd
import requests as request_lib
from databricks.agents.evals import judges, metric  # pylint: disable=ungrouped-imports
from mlflow import create_experiment, get_experiment_by_name, get_trace, start_run
from mlflow.evaluation import Assessment
from mlflow.models import infer_signature
from tqdm import tqdm

from databricks.kie.t2t_custom_model import CustomModel, ModelWrapper
from databricks.kie.t2t_instruction_analyzer import InstructionInfo, generate_instruction_info
from databricks.kie.t2t_metrics import create_classic_metrics
from databricks.kie.t2t_schema import SideBySideResult, T2TSystemParams, T2TSystemResult, Text2TextTaskSpec
from databricks.kie.t2t_utils import (INSTRUCTION_AND_INPUT_PROMPT_TEMPLATE, LLM_JUDGE_EVAL_PROMPT_TEMPLATE,
                                      generate_evaluation_criteria, validate_secret_exists)
from databricks.kie.task_spec import get_default_experiment_name_with_suffix
from databricks.model_training.api.utils import get_host_and_token_from_env
from databricks.model_training.mlflow_utils import get_mlflow_client
from databricks.model_training.version import __version__


class Text2TextRunner:
    """
    A class to manage the Text2Text model lifecycle, including training, evaluation, and deployment.

    Main API:
    - compile(): Optimizes prompt / model given provided hyperparameters.
    - update_system_param(): Updates the state of the runner with new instruction, evaluation criteria, or example data.
    - evaluate_models(): Evaluates all models in the runner.
    - visualize_models(): Visualizes model predictions for a subset of test data.
    - deploy_model(): Deploys a model to a production endpoint.
    """
    NUM_FEW_SHOTS = [0, 5]
    MIN_EVAL_DATA = 5
    MAX_EVAL_DATA = 100
    MAX_EXAMPLES_FOR_VISUALIZATION = 5
    DEFAULT_MODEL_ID = "gpt-4o-2024-08-06-text2text"
    MODEL_SELECTION = ["gpt-4o-2024-08-06-text2text", "gpt-4o-mini-2024-07-18-text2text"]
    MODEL_SELECTION_WITH_STRUCTURED_OUTPUT_SUPPORT = ["gpt-4o-2024-08-06-text2text", "gpt-4o-mini-2024-07-18-text2text"]
    INSTRUCTION_TEMPLATE = """
        INSTRUCTION: {optimized_instruction}
        OUTPUT FORMAT: {output_format}

        Given the following instruction and output format, provide response that follows the instruction and output format given the user input.
        User input will be provided below as part of the user message.  
    """

    def __init__(self, task_spec: Text2TextTaskSpec):
        """
        Initializes the Text2TextRunner with the provided parameters.

        Args:
            task_spec: Text2TextTaskSpec object containing the instruction, example data, and evaluation criteria
        """
        self.model_map: Dict[str, ModelWrapper] = {}
        self.test_data: List[Dict[str, str]] = []
        self.train_data: List[Dict[str, str]] = []
        self.instruction_info = None
        self.default_model: Optional[ModelWrapper] = None
        self.prev_task_spec: Optional[Text2TextTaskSpec] = None
        self.prev_params: Optional[T2TSystemParams] = None
        self.custom_metrics = []

        self.current_task_spec = task_spec
        self.current_params = task_spec.to_system_param()
        experiment_name = task_spec.experiment_name
        if not experiment_name:
            now = datetime.now()
            formatted_string = now.strftime("%Y_%m_%d_%H_%M_%S")
            experiment_name = get_default_experiment_name_with_suffix(f"t2t_{formatted_string}")
            print(f"Experiment name not provided. Creating new experiment: {experiment_name}")

        self.experiment_id = self._get_exp_id(experiment_name)
        self.mlflow_client = get_mlflow_client()

    def compile(self) -> T2TSystemResult:
        self._prepare_data()

        # get common data shared between previous state and new state.
        common_data = []
        prev_res = []
        if self.prev_params:
            print("Generating response for previous state for side by side comparison...")
            # generate req/response for previous state. This will be used to compare with current state.
            common_data = self.prev_params.get_sxs_data(self.current_params)[0:self.MAX_EXAMPLES_FOR_VISUALIZATION]
            assert self.default_model is not None
            prev_res = self._visualize_model(self.default_model, data=common_data)

        print("Compiling new system...")
        # Construct evaluation criteria. Infer from examples if not provided.
        if not self.current_params.evaluation_criteria:
            print("Inferring evaluation criteria automatically...")
            inferred_eval_criteria = generate_evaluation_criteria(self.current_params.instruction, self.train_data)
            self.current_params.evaluation_criteria = inferred_eval_criteria
            self.current_task_spec.evaluation_criteria = inferred_eval_criteria

        self._create_custom_metrics()

        # update models & system state to create new Text2Text models
        num_icl_examples = min(len(self.train_data), 10)
        self.instruction_info: Optional[InstructionInfo] = generate_instruction_info(
            self.train_data[:num_icl_examples], self.current_params.instruction,
            self.current_params.evaluation_criteria)

        self._create_classic_metrics()
        self.model_map = {}
        self._create_models()

        # generate response for common data for current state -- to be used for SxS comparison.
        current_res = []
        if prev_res:
            assert common_data and self.default_model is not None
            print("Getting new system response for side by side comparison..")
            current_res = self._visualize_model(self.default_model, data=common_data)

        return self._get_model_result(prev_res, current_res)

    def _get_judge_prompt(self, eval_criteria: str) -> str:
        return LLM_JUDGE_EVAL_PROMPT_TEMPLATE.format(
            eval_criteria=eval_criteria
        ) + "\n The instruction & request is : '{request}' \n  Model Response is: '{response}'"

    def _get_model_result(self, prev_result, current_result) -> T2TSystemResult:
        sxs_results = None
        if self.prev_params:
            sxs_results = []
            assert prev_result and current_result
            for prev, current in zip(prev_result, current_result):
                sxs_results.append(
                    SideBySideResult(request=prev['input'],
                                     previous_response=prev['model_output'],
                                     current_response=current['model_output']))

        assert self.default_model is not None

        data = self.test_data[0:self.MAX_EXAMPLES_FOR_VISUALIZATION]
        print("Visualizing model results..")
        examples = self._visualize_model(self.default_model, data=data)

        print("Evaluating default model...")
        eval_result = self._evaluate_model(self.default_model, data=data)
        models = [model.run_name for model in self.model_map.values()]

        return T2TSystemResult(
            eval_result=eval_result,
            examples=examples,
            sxs_result=sxs_results,
            current_task_spec=self.current_task_spec,
            experiment_id=self.experiment_id,
            model_candidates=models,
            default_model=self.default_model.run_name,
            instruction_info=self.instruction_info,
        )

    def get_system_param(self):
        models = list(self.model_map.keys())

        return {
            "instruction": self.current_params.instruction,
            "eval_criteria": self.current_task_spec.evaluation_criteria,
            "example_data": self.test_data[:self.MAX_EXAMPLES_FOR_VISUALIZATION],
            "experiment_id": self.experiment_id,
            "model_candidates": models,
            "instruction_info": self.instruction_info,
        }

    def update_system_param(self, new_spec: Text2TextTaskSpec):
        """Updates the state of the runner with new Text2Text spec."""
        if new_spec.experiment_name is not self.current_task_spec.experiment_name:
            raise ValueError("Experiment name cannot be updated.")

        new_system_param = new_spec.to_system_param()
        self.prev_params = self.current_params
        self.prev_task_spec = self.current_task_spec
        self.current_task_spec = new_spec
        self.current_params = new_system_param
        return self.compile()

    def evaluate_models(self):
        """Evaluates all models in the runner using the provided evaluation criteria."""
        result = {}
        if not self.current_task_spec.evaluation_criteria:
            raise ValueError("Evaluation criteria is not provided. Make sure to call `compile()` API.")

        for model_wrapper in self.model_map.values():
            model_eval = self._evaluate_model(model_wrapper)
            result[model_wrapper.run_name] = model_eval
        return result

    def visualize_models(self):
        """Runs a subset of test data through each model and returns the results."""
        result = {}
        for model_wrapper in self.model_map.values():
            result[model_wrapper.run_name] = self._visualize_model(model_wrapper)
        return pd.DataFrame.from_dict(result)

    def deploy_model(self, run_name: str, schema_name: str, endpoint_name: str, secret_scope: str, secret_key: str):
        """Deploys a model to a production endpoint using the provided secret.

        Args:
            run_name: The run name of the model to deploy.
            schema_name: The unity catalog schema name to use for the model.
            endpoint_name: The name of the endpoint to deploy.
            secret_scope: The secret scope containing the secret.
            secret_key: The secret key to use for the deployment.
        """
        if not validate_secret_exists(secret_scope, secret_key):
            raise ValueError(f"Secret {secret_key} in scope {secret_scope} does not exist.")

        model_wrapper = None
        model_wrapper = self.model_map.get(run_name, None)
        if not model_wrapper:
            raise ValueError(
                f"invalid model descriptor.. input: `{run_name}` candidates: {list(self.model_map.keys())}")

        self._register_model(model_wrapper, schema_name)

        model_url = model_wrapper.get_model_path()
        production_model = mlflow.pyfunc.load_model(model_url)
        api_url, api_token = get_host_and_token_from_env()
        data = {
            "name": endpoint_name,
            "config": {
                "served_entities": [{
                    "entity_name": model_wrapper.registered_model_name,
                    "entity_version": '1',
                    "workload_size": "Small",
                    "scale_to_zero_enabled": True,
                    "environment_vars": {
                        "DATABRICKS_HOST": api_url,
                        "DATABRICKS_TOKEN": "{{" + f"secrets/{secret_scope}/{secret_key}" + "}}",
                    }
                }]
            }
        }

        headers = {"Context-Type": "text/json", "Authorization": f"Bearer {api_token}"}

        response = request_lib.post(url=f"{api_url}/api/2.0/serving-endpoints", json=data, headers=headers, timeout=60)
        print("Deployed endpoint.. Response status:", response.status_code)
        print("Deployed endpoint.. Response text:", response.text, "\n")
        return production_model

    def _get_exp_id(self, experiment_name: str):
        experiment = get_experiment_by_name(experiment_name)
        if not experiment:
            return create_experiment(name=experiment_name)
        return experiment.experiment_id

    def _create_custom_metrics(self):
        """Create custom metrics for the model.

        Generates custom metrics for the model based on the self.current_task_spec.eval_criteria.
        Assumes that self.evaluation_criteria is already set.(i.e. compile() has been called)
        """

        def _wrapper_fn(request: str, response: str, guideline: str, metric_name: str):
            assessment = judges.guideline_adherence(
                request=request,
                response=response,
                guidelines=[guideline],
            )
            return Assessment(name=metric_name,
                              value=assessment.value,
                              rationale=assessment.rationale,
                              source=assessment.source)

        eval_criteria = self.current_task_spec.evaluation_criteria
        assert eval_criteria
        self.custom_metrics = []
        for criterion in eval_criteria:
            metric_name = re.sub(r'[^a-zA-Z0-9]', '_', criterion)
            custom_metric_fn = functools.partial(_wrapper_fn, guideline=criterion, metric_name=metric_name)
            custom_metric = metric(eval_fn=custom_metric_fn, name=metric_name)
            self.custom_metrics.append(custom_metric)

    def _create_classic_metrics(self):
        """Creates classic metrics for given the instruction type. Assumes self.instruction_info is already set."""
        assert self.instruction_info.instruction_type
        classic_metrics = create_classic_metrics(self.instruction_info.instruction_type)
        self.custom_metrics.extend(classic_metrics)

    def _prepare_data(self):
        """Randomly sample half of example_data to test data and other half to train data."""
        # TODO: support unlabeled data / output generation.
        assert self.current_params.labelled_examples
        data = self.current_params.labelled_examples
        example_data = [{'request': datum[0], 'response': datum[1]} for datum in data]
        example_data_size = len(data)
        if example_data_size < self.MIN_EVAL_DATA:
            self.test_data = example_data
            return

        indices = list(range(example_data_size))
        random.shuffle(indices)
        split_index = example_data_size // 2

        train_indices = indices[:split_index]
        test_indices = indices[split_index:]

        self.train_data = [example_data[i] for i in train_indices]
        self.test_data = [example_data[i] for i in test_indices]

    def _create_optimized_instruction(self):
        if not self.instruction_info:
            raise ValueError("Instruction info is not provided. Make sure to call `compile()` API.")

        optimized_instruction = self.instruction_info.optimized_instruction
        output_format = self.instruction_info.output_format

        return self.INSTRUCTION_TEMPLATE.format(optimized_instruction=optimized_instruction,
                                                output_format=output_format)

    def _create_models(self):
        for model_id in self.MODEL_SELECTION:
            if model_id in self.MODEL_SELECTION_WITH_STRUCTURED_OUTPUT_SUPPORT:
                model = self._create_cot_model(model_id)
                self.model_map[model.run_name] = model
                if model_id == self.DEFAULT_MODEL_ID:
                    self.default_model = model

            model = self._create_baseline_model(model_id)
            self.model_map[model.run_name] = model

            for n in self.NUM_FEW_SHOTS:
                if n > len(self.train_data):
                    break

                model = self._create_text2text_model(model_id, self._create_optimized_instruction(), num_few_shot=n)
                self.model_map[model.run_name] = model

    def _create_cot_model(self, model_id: str):
        num_few_shot_examples = max(n for n in self.NUM_FEW_SHOTS if n <= len(self.train_data))
        model_descriptor = f"{model_id}_chain_of_thought"
        return self._create_text2text_model(model_id,
                                            self._create_optimized_instruction(),
                                            num_few_shot=num_few_shot_examples,
                                            use_cot=True,
                                            model_descriptor=model_descriptor)

    def _create_baseline_model(self, model_id: str):
        model_descriptor = f"{model_id}_baseline"
        return self._create_text2text_model(model_id,
                                            self.current_params.instruction,
                                            num_few_shot=0,
                                            use_cot=False,
                                            model_descriptor=model_descriptor)

    def _register_model(self, model_wrapper: ModelWrapper, schema_path: str):
        custom_model = model_wrapper.custom_model
        run_id = model_wrapper.run_id
        registered_model_name = f'{schema_path}.t2t_{self.experiment_id}_{run_id}'
        model_wrapper.registered_model_name = registered_model_name
        with start_run(experiment_id=self.experiment_id, run_id=run_id):
            mlflow.pyfunc.log_model(
                "model",
                python_model=custom_model,
                pip_requirements=["pandas", "mlflow", f"databricks-genai=={__version__}", "databricks-connect"],
                signature=custom_model.signature,
                registered_model_name=registered_model_name)

    def _create_text2text_model(
        self,
        model_id: str,
        instruction: str,
        num_few_shot: int = 0,
        use_cot: bool = False,
        model_descriptor: Optional[str] = None,
    ) -> ModelWrapper:
        description = (f't2t_model_registry_{model_descriptor}'
                       if model_descriptor else f't2t_model_registry_{model_id}_few_shot_{num_few_shot}')
        with start_run(
                experiment_id=self.experiment_id,
                description=description,
        ) as run:
            few_shot_examples = []
            if num_few_shot:
                few_shot_examples = random.sample(self.train_data, num_few_shot)

            signature = infer_signature(model_input=["input"], model_output=["output"])

            custom_model = CustomModel(
                instruction,
                signature,
                model_id,
                few_shot_examples,
                model_descriptor=model_descriptor,
                use_cot=use_cot,
            )
            run_id = run.info.run_id
            run_name = run.info.run_name
            return ModelWrapper(custom_model, run_id, run_name, registered_model_name=None)

    def _visualize_model(self, model_wrapper: ModelWrapper, data: Optional[List[Dict[str, str]]] = None):
        custom_model = model_wrapper.custom_model
        model_results = []
        eval_data = data or self.test_data[0:self.MAX_EXAMPLES_FOR_VISUALIZATION]
        for datum in tqdm(eval_data, desc=f"Visualizing model {model_wrapper.run_name}"):
            req = datum['request']
            output_dict = {}
            output_dict["input"] = req
            output_dict["model_output"] = custom_model.predict(None, req)
            if 'response' in datum:
                output_dict["label"] = datum['response']
            model_results.append(output_dict)
        return model_results

    def _evaluate_model(self,
                        model_wrapper: ModelWrapper,
                        data: Optional[List[Dict[str, str]]] = None) -> mlflow.models.EvaluationResult:
        eval_data = data or self.test_data[0:self.MAX_EVAL_DATA]
        if not self.current_params.evaluation_criteria:
            raise ValueError("Evaluation criteria is not provided. Please invoke `compile()` API.")

        custom_model = model_wrapper.custom_model
        requests = [
            INSTRUCTION_AND_INPUT_PROMPT_TEMPLATE.format(instruction=self.current_params.instruction,
                                                         inp=datum['request']) for datum in eval_data
        ]
        ground_truth = [datum['response'] for datum in eval_data]

        with start_run(experiment_id=self.experiment_id, run_id=model_wrapper.run_id):
            model_responses = []
            traces = []
            for req in tqdm(requests, desc=f"Evaluating model {model_wrapper.run_name}"):
                root_span = self.mlflow_client.start_trace(name=model_wrapper.run_name)
                request_id = root_span.request_id
                model_response = custom_model.predict(None, req)
                self.mlflow_client.end_trace(request_id)

                traces.append(get_trace(request_id).to_json())
                model_responses.append(model_response)

            eval_df = pd.DataFrame({
                "request": requests,
                "expected_response": ground_truth,
                "response": model_responses,
                "trace": traces
            })

            return mlflow.evaluate(
                data=eval_df,
                model_type="databricks-agent",  # Enable Mosaic AI Agent Evaluation
                extra_metrics=self.custom_metrics,
                # Disable built-in judges.
                evaluator_config={'databricks-agent': {
                    "metrics": [],
                }})



"""Utility functions for text processing."""

import difflib
import re

__all__ = [
    'clean_text',
    'normalize_text',
    'fuzzy_string_match',
]


def clean_text(text: str) -> str:
    # Replace common character that shows up in PDF parsed text
    text = text.replace('\xa0', ' ')

    # Strip off leading and trailing white spaces
    text = text.strip()

    # Lower case
    text = text.lower()

    # Replace consecutive white spaces with single ones
    text = re.sub(r'\s+', ' ', text)

    return text


def normalize_text(text: str) -> str:
    # Apply cleaning
    text = clean_text(text)

    # Remove punctuation
    text = re.sub(r'[^\w\s]', '', text)

    # Remove ordinal suffixes
    text = re.sub(r'\b(\d+)(st|nd|rd|th)\b', lambda x: x.group(1), text)

    return text


def fuzzy_string_match(
    text_1: str,
    text_2: str,
    long_ratio_threshold: float = 0.7,
    short_ratio_threshold: float = 0.5,
    similarity_threshold: float = 0.8,
    min_length_threshold: int = 3,
) -> bool:
    """Check if two strings are similar based on a combination of string similarity and ratio of longest match.

    Args:
        text_1 (str): The first text to compare.
        text_2 (str): The second text to compare.
        long_ratio_threshold (float, optional): Minimum threshold for the larger ratio of longest common substring
            to string length. Defaults to 0.7.
        short_ratio_threshold (float, optional): Minimum threshold for the smaller ratio of longest common substring
            to string length. Defaults to 0.5.
        similarity_threshold (float, optional): Minimum overall sequence similarity ratio required for strings
            to be considered matching. Defaults to 0.8.
        min_length_threshold (int, optional): Minimum length of input strings for fuzzy matching.
            Strings shorter than this will be compared exactly. Defaults to 3.

    Returns:
        bool: True if the strings are similar, False otherwise.
    """
    if len(text_1) <= min_length_threshold or len(text_2) <= min_length_threshold:
        return clean_text(text_1) == clean_text(text_2)

    sm = difflib.SequenceMatcher(None, clean_text(str(text_1)), clean_text(str(text_2)))
    similarity = sm.ratio()
    longest_match = sm.find_longest_match()
    ratio_a = longest_match.size / len(sm.a)  # type: ignore
    ratio_b = longest_match.size / len(sm.b)  # type: ignore
    longest_match_ratio = max(ratio_a, ratio_b)
    shortest_match_ratio = min(ratio_a, ratio_b)

    is_ratio_match = longest_match_ratio > long_ratio_threshold and shortest_match_ratio > short_ratio_threshold
    is_similarity_match = similarity > similarity_threshold

    return is_ratio_match or is_similarity_match



"""mlflow utility functions for handling runs and metadata"""
import json
from typing import Dict, NamedTuple, Optional

from databricks.sdk import WorkspaceClient  # pylint: disable=ungrouped-imports
from mlflow import MlflowClient
from mlflow.data.delta_dataset_source import DeltaDatasetSource
from mlflow.data.meta_dataset import MetaDataset
from mlflow.entities import DatasetInput, InputTag, SourceType
from mlflow.utils.mlflow_tags import (MLFLOW_DATABRICKS_NOTEBOOK_ID, MLFLOW_DATABRICKS_NOTEBOOK_PATH,
                                      MLFLOW_DATASET_CONTEXT, MLFLOW_SOURCE_NAME, MLFLOW_SOURCE_TYPE)

from databricks.model_training.api.utils import get_current_notebook_details


def get_mlflow_client() -> MlflowClient:
    return MlflowClient(tracking_uri="databricks", registry_uri="databricks-uc")


class DeltaTableDetails(NamedTuple):
    catalog: Optional[str]
    name: Optional[str]
    full_name: Optional[str]
    version: Optional[int]


def get_delta_table_details(path: str) -> DeltaTableDetails:
    w = WorkspaceClient()
    details = w.tables.get(path, include_delta_metadata=True)
    version = None
    if details.delta_runtime_properties_kvpairs:
        if details.delta_runtime_properties_kvpairs.delta_runtime_properties:
            attributes = details.delta_runtime_properties_kvpairs.delta_runtime_properties.get("commitAttributes")
            if attributes:
                version = json.loads(attributes)["version"]

    return DeltaTableDetails(catalog=details.catalog_name,
                             name=details.name,
                             full_name=details.full_name,
                             version=version)


def log_delta_table_source(run_id: str, path: str, name: str = "Dataset", split: Optional[str] = None):
    client = get_mlflow_client()

    # Get the source delta table
    table_details = get_delta_table_details(path)
    source = DeltaDatasetSource(delta_table_name=path, delta_table_version=table_details.version)
    #TODO: Could add schema using the columns field from get_delta_table_details
    meta = MetaDataset(source, name=name)

    # Contstruct the entities
    tags = [InputTag(key=MLFLOW_DATASET_CONTEXT, value=split)] if split else []
    inputs = [DatasetInput(dataset=meta._to_mlflow_entity(), tags=tags)]  # pylint: disable=protected-access

    return client.log_inputs(run_id=run_id, datasets=inputs)


def update_run_tags(run_id: str, tags: Dict[str, str]):
    client = get_mlflow_client()
    for k, v in tags.items():
        client.set_tag(run_id, k, v)


def change_run_name(run_id: str, new_name: str) -> None:
    client = get_mlflow_client()
    return client.update_run(run_id=run_id, name=new_name)


def add_notebook_source(run_id: str):
    details = get_current_notebook_details()
    if not details:
        return
    tags = {
        MLFLOW_SOURCE_NAME: details.notebook_path,
        MLFLOW_SOURCE_TYPE: SourceType.to_string(SourceType.NOTEBOOK),
        MLFLOW_DATABRICKS_NOTEBOOK_PATH: details.notebook_path,
        MLFLOW_DATABRICKS_NOTEBOOK_ID: details.notebook_id,
    }
    update_run_tags(run_id, tags)



def test_package():
    import databricks
    del databricks

    from databricks import model_training
    assert model_training
    del model_training

    from databricks.model_training import exceptions
    assert exceptions
    del exceptions

    from databricks.model_training import types
    assert types
    assert types.training_run
    del types



"""List multiple model training runs"""

from datetime import datetime
from typing import List, Optional, Union

from databricks.model_training.api.engine import get_return_response, run_paginated_mapi_request
from databricks.model_training.api.validation import parse_datetime
from databricks.model_training.types import TrainingRun
from databricks.model_training.types.common import ObjectList
from databricks.model_training.types.run_status import RunStatus

DEFAULT_LIMIT = 100

QUERY_FUNCTION = 'getFinetunesPaginated'
VARIABLE_DATA_NAME = 'getFinetunesPaginatedData'

QUERY = f"""
query GetFinetunesPaginated(${VARIABLE_DATA_NAME}: GetFinetunesPaginatedInput!) {{
  {QUERY_FUNCTION}({VARIABLE_DATA_NAME}: ${VARIABLE_DATA_NAME}) {{
    cursor
    hasNextPage
    finetunes {{
        id
        name
        status
        createdById
        createdByEmail
        createdAt
        updatedAt
        startedAt
        completedAt
        reason
        estimatedEndTime
        isDeleted
    }}
  }}
}}"""

QUERY_WITH_DETAILS_DB = f"""
query GetFinetunesPaginated(${VARIABLE_DATA_NAME}: GetFinetunesPaginatedInput!) {{
  {QUERY_FUNCTION}({VARIABLE_DATA_NAME}: ${VARIABLE_DATA_NAME}) {{
    cursor
    hasNextPage
    finetunes {{
        id
        name
        status
        createdById
        createdByEmail
        createdAt
        updatedAt
        startedAt
        completedAt
        reason
        estimatedEndTime
        isDeleted
        details {{
            model
            taskType
            trainDataPath
            saveFolder
            evalDataPath
            evalPrompts
            trainingDuration
            learningRate
            contextLength
            experimentTracker
            customWeightsPath
            dataPrepConfig
            formattedFinetuningEvents {{
                eventType
                eventTime
                eventMessage
            }}
        }}
    }}
  }}
}}"""


def list(  # pylint: disable=redefined-builtin
    training_runs: Optional[Union[List[str], List[TrainingRun], ObjectList[TrainingRun]]] = None,
    experiment_id: Optional[str] = None,
    *,
    statuses: Optional[Union[List[str], List[RunStatus]]] = None,
    user_emails: Optional[List[str]] = None,
    before: Optional[Union[str, datetime]] = None,
    after: Optional[Union[str, datetime]] = None,
    include_details: bool = True,
    limit: Optional[int] = DEFAULT_LIMIT,
) -> ObjectList[TrainingRun]:
    """List training runs

    Args:
        training_runs (Optional[Union[List[str], List[TrainingRun], ObjectList[TrainingRun]]], optional):
        The training runs to list. Defaults to None.
        experiment_id (Optional[str], optional): The MLflow experiment ID to filter by. Defaults to None.
        statuses (Optional[Union[List[str], List[RunStatus]], optional): The statuses to filter by. Defaults to None.
        user_emails (Optional[List[str]], optional): The user emails to filter by. Defaults to None.
        include_details (bool, optional): Whether to include details in the response. Defaults to True.
        before (Optional[Union[str, datetime]], optional): The date to filter before. Defaults to None.
        after (Optional[Union[str, datetime]], optional): The date to filter after. Defaults to None.
        limit (Optional[int], optional): The maximum number of runs to return. Defaults to the 100 most recent.

    Returns:
        ObjectList[TrainingRun]: A list of training runs
    """
    filters = {}
    if training_runs:
        filters['name'] = {'in': [r.name if isinstance(r, TrainingRun) else r for r in training_runs]}
    if experiment_id:
        filters['mlflowExperimentId'] = {'equals': experiment_id}
    if statuses:
        filters['status'] = {'in': [s.value if isinstance(s, RunStatus) else s for s in statuses]}
    if before or after:
        date_filters = {}
        if before:
            date_filters['lt'] = before.astimezone().isoformat() if isinstance(before,
                                                                               datetime) else parse_datetime(before)
        if after:
            date_filters['gte'] = after.astimezone().isoformat() if isinstance(after,
                                                                               datetime) else parse_datetime(after)
        filters['createdAt'] = date_filters

    variables = {
        VARIABLE_DATA_NAME: {
            'filters': filters,
            'includeDeleted': False,
            'limit': limit,
        },
    }

    if user_emails:
        if variables[VARIABLE_DATA_NAME].get('entity'):
            variables[VARIABLE_DATA_NAME]['entity']['emails'] = user_emails
        else:
            variables[VARIABLE_DATA_NAME]['entity'] = {'emails': user_emails}

    response = run_paginated_mapi_request(
        query=QUERY_WITH_DETAILS_DB if include_details else QUERY,
        query_function=QUERY_FUNCTION,
        return_model_type=TrainingRun,
        variables=variables,
    )
    return get_return_response(response)



"""List multiple model training runs"""

from databricks.model_training.api.engine import get_return_response, run_singular_mapi_request
from databricks.model_training.types import TrainingConstants, TrainingModelConstants
from databricks.model_training.types.common import ObjectList, ObjectType

QUERY_FUNCTION = 'getFinetuneConstants'

QUERY = f"""
query Models {{
  {QUERY_FUNCTION} {{
    models {{
      name
      displayName
      maxContextLength
    }}
  }}
}}"""


def get_models() -> ObjectList[TrainingModelConstants]:
    """List available models (for the Foundation Model Training API) and their associated info

    Returns:
        ObjectList[ModelInfo]: A list of models and their associated info
    """
    response = run_singular_mapi_request(query=QUERY,
                                         query_function=QUERY_FUNCTION,
                                         return_model_type=TrainingConstants,
                                         variables={})

    return ObjectList(get_return_response(response).models, ObjectType.TRAINING_MODEL_CONSTANTS)



"""Eval Utils for Tiles"""
from enum import Enum
from typing import Union

from databricks.agents.evals import judges
from mlflow.evaluation import Assessment

from databricks.kie.text_utils import fuzzy_string_match, normalize_text  # pylint: disable=ungrouped-imports

PrimitiveTypes = Union[int, float, str, bool, Enum]


# TODO: Needs to match TaskSpec definition of a metric
def exact_match(query: str, ground_truth: PrimitiveTypes, prediction: PrimitiveTypes) -> Assessment:
    del query  # unused
    result = ground_truth == prediction

    if result:
        rationale = 'Values matched exactly'
    else:
        rationale = 'Values did not match exactly'

    return Assessment(name='exact_match', value=result, rationale=rationale)


def normalized_match(query: str, ground_truth: PrimitiveTypes, prediction: PrimitiveTypes) -> Assessment:
    if not isinstance(ground_truth, str) or not isinstance(prediction, str):
        result = exact_match(query, ground_truth, prediction)
    else:
        gt_normalized = normalize_text(ground_truth)
        pred_normalized = normalize_text(prediction)
        result = exact_match(query, gt_normalized, pred_normalized)

    if result.value:
        rationale = 'Normalized values matched'
    else:
        rationale = 'Normalized values did not match'

    return Assessment(name='normalized_match', value=result.value, rationale=rationale)


def fuzzy_match(query: str, ground_truth: PrimitiveTypes, prediction: PrimitiveTypes) -> Assessment:
    if not isinstance(ground_truth, str) or not isinstance(prediction, str):
        result = exact_match(query, ground_truth, prediction).value
    else:
        result = fuzzy_string_match(ground_truth, prediction)

    if result:
        rationale = 'Values approximately matched'
    else:
        rationale = 'Values did not approximately match'
    return Assessment(name='fuzzy_match', value=result, rationale=rationale)


# TODO: Possibly need retries and robustness here
def correctness(query: str, ground_truth: PrimitiveTypes, prediction: PrimitiveTypes) -> Assessment:
    if not isinstance(ground_truth, str) or not isinstance(prediction, str):
        return exact_match(query, ground_truth, prediction)

    # TODO: I think this supports guidelines now and should be updated to use that
    correctness_result = judges.correctness(request=query, response=prediction, expected_response=ground_truth)

    # Correctness judge does not return lists
    assert isinstance(correctness_result, Assessment)

    if correctness_result.error_code is not None:
        return fuzzy_match(query, ground_truth, prediction)

    bool_value = correctness_result.value == 'yes'

    remade_result = Assessment(
        name='correctness_judge',
        value=bool_value,
        rationale=correctness_result.rationale,
    )
    return remade_result



"""List multiple model checkpoint paths"""

import os
from typing import List

import mlflow

from databricks.model_training.api.exceptions import DatabricksGenAIResponseError
from databricks.model_training.types import TrainingRun

COMPOSER_CKPT_PREFIX = 'ep'


def get_checkpoints(training_run: TrainingRun) -> List[str]:
    """List MLflow checkpoint paths for the MosaicAI Model Training SDK

    Args:
        training_run (TrainingRun): The training run to get checkpoints for

    Returns:
        List[str]: A list of checkpoint paths for that training run
    """

    if training_run is None:
        raise DatabricksGenAIResponseError('Must provide training run object to retrieve checkpoints')

    try:
        mlflow_run_id = training_run.run_id
        artifact_path = f'{training_run.name}/checkpoints'
        artifacts = mlflow.artifacts.list_artifacts(run_id=mlflow_run_id, artifact_path=artifact_path)

    except Exception as e:
        raise DatabricksGenAIResponseError(f'Checkpoints could not be retrieved for {training_run}') from e

    # Convert FileInfo objects to a list of paths (Composer checkpoints only)
    paths = [f.path for f in artifacts if os.path.basename(f.path).startswith(COMPOSER_CKPT_PREFIX)]
    # Join relative paths with the save_folder path to get the full checkpoint paths
    # TODO: Once MCLOUD-4934 is addressed, the following line can be removed
    training_run.save_folder = (training_run.save_folder.format(mlflow_experiment_id=training_run.experiment_id,
                                                                mlflow_run_id=training_run.run_id))
    paths = [os.path.join(training_run.save_folder, path) for path in paths]

    if len(paths) == 0:
        print('No checkpoints created yet for this run. Please try again later.')

    return paths



"""Prompt builder using Pydantic schema + examples"""
import json
from textwrap import dedent
from typing import Dict, List, Optional, Tuple

from pydantic import BaseModel


class PromptBuilder:
    """Builds an LLM prompt based on a predefined output schema given by a Pydantic model
    """

    def __init__(self, schema_model: type[BaseModel]):
        """
        Initialize the prompt builder with a Pydantic model and optional examples.
        
        Args:
            schema_model: The Pydantic model class defining the expected structure
            examples: Optional list of validated examples conforming to the schema
        """
        self.schema_model = schema_model

    def get_schema_description(self) -> str:
        """Generate a human-readable description of the schema requirements."""
        schema = self.schema_model.model_json_schema()

        def resolve_ref(ref: str, schema: dict) -> dict:
            """Resolve $ref by looking up in definitions."""
            if not ref.startswith('#/$defs/'):
                return {}
            ref_name = ref[len('#/$defs/'):]
            return schema.get('$defs', {}).get(ref_name, {})

        def format_field(name: str, field_schema: dict, schema: dict, indent: int = 0) -> List[str]:
            """Format a field for the prompt"""
            lines = []

            description = field_schema.get('description', '')
            required = name in schema.get('required', [])

            # Resolve $ref if present
            if '$ref' in field_schema:
                field_schema = resolve_ref(field_schema['$ref'], schema)
                field_schema.update(description=description, required=required)

            if 'type' in field_schema:
                field_type = field_schema.get('type', 'any')
            elif 'anyOf' in field_schema:
                any_of = field_schema['anyOf']
                any_of_non_null = [p for p in any_of if p.get('type') != 'null']

                if len(any_of_non_null) > 1:
                    raise ValueError("Union types are not supported.")

                if len(any_of_non_null) == 0:
                    raise RuntimeError("Unexpected error. No non-null type found.")

                any_of_non_null_item = any_of_non_null[0]
                if '$ref' in any_of_non_null_item:
                    any_of_non_null_item = resolve_ref(any_of_non_null_item['$ref'], schema)
                    any_of_non_null_item.update(description=description, required=required)

                field_type = any_of_non_null_item.get('type', 'any')
                field_schema = any_of_non_null_item
            else:
                field_type = 'any'

            if field_type == 'array':
                items = field_schema.get('items', {})
                if '$ref' in items:
                    items = resolve_ref(items['$ref'], schema)
                    field_schema = items
                field_type = f"{field_type}[{items.get('type', 'any')}]"

            # Build base line
            line = "  " * indent + f"- {name} ({field_type})"
            if description:
                line += f": {description}"
            if required:
                line += " (non-null)"
            else:
                line += " (nullable)"
            lines.append(line)

            # Handle nested objects
            if field_type in {'object', 'array[object]'} and 'properties' in field_schema:
                for nested_name, nested_schema in field_schema['properties'].items():
                    lines.extend(format_field(nested_name, nested_schema, schema, indent + 1))

            return lines

        fields = []
        for name, props in schema.get('properties', {}).items():
            fields.extend(format_field(name, props, schema))

        return "\n".join(fields)

    def build_prompt(self, examples: Optional[List[Tuple[str, Dict]]] = None, include_markdown: bool = False) -> str:
        """
        Build the instruction prompt with schema requirements and examples.
        
        Args:
            examples: Optional list of (input_text, output_json) pairs showing example extractions
        """
        include_markdown_instructions = """
All outputs must be formatted as valid JSON. Wrap the output in a markdown JSON code block.

Example of CORRECT output format:
```json
\\{"field1": "value1", "field2": null\\}
```

Example of INCORRECT output format:
\\{"field1": "value1", "field2": null\\}

""" if include_markdown else """
All outputs must be formatted as valid JSON. DO NOT include any markdown formatting.

Example of CORRECT output format:
\\{"field1": "value1", "field2": null\\}

Example of INCORRECT output format:
```json
\\{"field1": "value1", "field2": null\\}
```
"""

        def format_as_markdown(text: str) -> str:
            if include_markdown:
                return f"```json\n{text}\n```"
            return text

        prompt = dedent(f"""
Extract information according to this schema:

Required Output Structure:
{self.get_schema_description()}

Your task:
1. Extract the relevant information from the provided text
2. Format it exactly according to the schema above
3. Return the output ONLY as valid JSON. IMPORTANT NOTE: All double quotes must be escaped with a SINGLE backslash, not double backslash.
4. Ensure all fields are included
5. If a field is nullable, the information may not be present in the provided text. In this case, the value of the field should be null.
6. If a field is non-nullable, it MUST NOT be null
7. Use strings for text, numbers for numeric values, and booleans (true/false) as specified
{include_markdown_instructions}
""").strip()

        # Add input-output examples if provided
        if examples:
            example_texts = []
            for i, (input_text, output_json) in enumerate(examples[:3], 1):  # Limit to 3 examples
                # Ensure proper JSON formatting of output and escape quotes
                formatted_output = format_as_markdown(json.dumps(output_json, indent=2))
                # Escape single quotes in input text
                example_text = dedent(f"""
Example {i} Input:
{input_text}

Example {i} Output:
{formatted_output}
""").strip()
                example_texts.append(example_text)

            prompt += "\n\nExtraction Examples:\n" + "\n\n".join(example_texts)

        return prompt.replace("'", "\\'").replace('"', '\\"')



from unittest.mock import Mock, patch

import pytest

from databricks.model_serving.utils import get_latest_model_version, get_model_specs


class MockModelVersion:

    def __init__(self, version: str):
        self.version = version


def test_get_latest_model_version_single_version():
    # Arrange
    model_name = 'test_model'
    mock_version = MockModelVersion('1')

    mock_client = Mock()
    mock_client.search_model_versions.return_value = [mock_version]

    with patch('databricks.model_serving.utils.MlflowClient', return_value=mock_client):
        result = get_latest_model_version(model_name)

    # Assert
    mock_client.search_model_versions.assert_called_once_with("name='test_model'")
    assert result == '1'


def test_get_latest_model_version_multiple_versions():
    # Arrange
    model_name = 'test_model'
    mock_versions = [MockModelVersion('1'), MockModelVersion('2'), MockModelVersion('3')]
    mock_client = Mock()
    mock_client.search_model_versions.return_value = mock_versions

    with patch('databricks.model_serving.utils.MlflowClient', return_value=mock_client):
        result = get_latest_model_version(model_name)

    # Assert
    mock_client.search_model_versions.assert_called_once_with("name='test_model'")
    assert result == '3'


def test_get_latest_model_version_unordered_versions():
    # Arrange
    model_name = 'test_model'
    mock_versions = [MockModelVersion('2'), MockModelVersion('1'), MockModelVersion('5'), MockModelVersion('3')]
    mock_client = Mock()
    mock_client.search_model_versions.return_value = mock_versions

    with patch('databricks.model_serving.utils.MlflowClient', return_value=mock_client):
        result = get_latest_model_version(model_name)

    # Assert
    mock_client.search_model_versions.assert_called_once_with("name='test_model'")
    assert result == '5'


def test_get_latest_model_version_empty_list():
    # Arrange
    model_name = 'test_model'
    mock_client = Mock()
    mock_client.search_model_versions.return_value = []

    with patch('databricks.model_serving.utils.MlflowClient', return_value=mock_client):
        with pytest.raises(ValueError):
            result = get_latest_model_version(model_name)


def test_all_model_specs_have_required_fields():
    # These fields may not apply to all models, but they
    # definitely apply to all "base" models that we want to use
    # TODO: We should also test for endpoint because it is required for the ground_truth models
    with patch('databricks.model_serving.utils.get_latest_model_version', return_value='1'):
        model_specs = get_model_specs()
        for model_spec in model_specs.values():
            assert model_spec.ft_model_name is not None
            assert model_spec.uc_schema is not None



"""Provisioned throughput endpoint for model serving"""

import json
import logging
import re
import threading
import time
from abc import abstractmethod
from concurrent.futures import ThreadPoolExecutor, as_completed
from contextlib import contextmanager
from enum import Enum
from functools import partial
from typing import Any, Generator, List, NamedTuple, Optional, Tuple, Type, TypeVar, cast

import requests
from mlflow.deployments import DatabricksDeploymentClient, DatabricksEndpoint, get_deploy_client
from pydantic import BaseModel
from requests import RequestException
from requests.exceptions import HTTPError
from rich.progress import BarColumn, Progress, TaskProgressColumn, TextColumn, TimeElapsedColumn, TimeRemainingColumn
from tenacity import retry, retry_if_exception, stop_after_attempt, wait_exponential

from databricks.model_serving.utils import get_latest_model_version
from databricks.model_training.api.utils import get_host_and_token_from_env, get_me, md5_hash
from databricks.model_training.logging import console

logger = logging.getLogger(__name__)

DEFAULT_MAX_RETRIES = 5
DEFAULT_BACKOFF_EXPONENTIAL = 1


class EndpointTask(str, Enum):
    COMPLETIONS = 'llm/v1/completions'
    CHAT = 'llm/v1/chat'


class EndpointClass(str, Enum):
    QUERY = '/serving-endpoints'
    BATCH = '/batch-inference'


class GenerationOutput(NamedTuple):
    response: str
    total_tokens: int


def should_retry_http(exception: BaseException) -> bool:
    """Only retry on specific HTTP status codes."""
    if isinstance(exception, HTTPError):
        # Don't retry on client errors except 429 (rate limit)
        if 400 <= exception.response.status_code < 500:
            return exception.response.status_code == 429
        # Retry on server errors
        return 500 <= exception.response.status_code < 600
    # Retry on connection errors, timeouts etc
    return isinstance(exception, RequestException)


T = TypeVar('T', bound='BaseEndpoint')


class BaseEndpoint:
    """Base endpoint class to query PayGo and PT endpoints
    """
    endpoint: DatabricksEndpoint
    client: DatabricksDeploymentClient
    tokens_processed: int = 0
    __endpoint_class: EndpointClass = EndpointClass.QUERY

    def __init__(self):
        dbx_client = get_deploy_client('databricks')
        if dbx_client is None:
            raise RuntimeError('Failed to get Databricks deploy client.')
        self.client = cast(DatabricksDeploymentClient, dbx_client)

        self._token_lock = threading.Lock()

        self.endpoint = self.setup_endpoint()

        if self.endpoint.get("task") and self.endpoint.get("task") not in set(EndpointTask):
            # This really shouldn't happen but just in case.
            raise ValueError(
                f'Unsupported task type configured for endpoint {self.endpoint.name}: {self.endpoint.task}')

    @abstractmethod
    def setup_endpoint(self) -> DatabricksEndpoint:
        """Perform any required setup and return an endpoint object from a call to client.get_endpoint()
        """

    def __str__(self):
        return f'endpoints:/{self.endpoint.name} ({self.endpoint.ready})'

    def __enter__(self: T) -> T:
        return self

    def __exit__(self, _exc_type, _exc_value, _traceback) -> None:
        del _exc_type, _exc_value, _traceback

    def _get_deployment_endpoint(self) -> str:
        assert self.endpoint.name is not None and isinstance(self.endpoint.name,
                                                             str)  # If self.endpoint is set, it'll have a str name
        return '/'.join([self.__endpoint_class.value, self.endpoint.name])

    def is_endpoint_ready(self) -> bool:
        """
        Returns True if the endpoint is scaled from zero and ready to serve
        requests. Returns False otherwise.
        """
        return self.endpoint.state.ready == 'READY' and self.endpoint.state.config_update == 'NOT_UPDATING'

    def _set_endpoint_class(self, endpoint_class: EndpointClass) -> None:
        self.__endpoint_class = endpoint_class

    def query(
        self,
        input_data: dict,
        max_retries: int = DEFAULT_MAX_RETRIES,
        backoff_exponential: float = DEFAULT_BACKOFF_EXPONENTIAL,
    ) -> GenerationOutput:
        """
        Queries the PT endpoint with the given input.

        Args:
            input_data (dict): Input data (or arguments) to pass to the deployment or model endpoint for inference.
        """

        @retry(retry=retry_if_exception(should_retry_http),
               stop=stop_after_attempt(max_retries),
               wait=wait_exponential(multiplier=backoff_exponential),
               reraise=True)
        def retry_query(input_data: dict) -> GenerationOutput:
            result = self.client.predict(
                endpoint=self._get_deployment_endpoint(),
                inputs=input_data,
            )
            if not result:
                raise RuntimeError(f'Failed to generate completion for input: {input_data}')

            if self.endpoint.get("task") == EndpointTask.COMPLETIONS:
                response = result['choices'][0]['text']
            else:
                response = result['choices'][0]['message']['content']

            total_tokens = result['usage']['total_tokens']
            return GenerationOutput(response=response, total_tokens=total_tokens)

        return retry_query(input_data)

    def generate_completion(self,
                            prompt: str,
                            system_prompt: Optional[str] = None,
                            response_format: Optional[Type[BaseModel]] = None,
                            temperature: float = 0.0,
                            max_tokens: int = 512) -> GenerationOutput:
        """
        Uses PT endpoint to generate completion for a given prompt string.

        Args:
            prompt (str): The prompt to generate completion for.
            temperature (float): The temperature to use for sampling. Default is 1.0.
            max_tokens (int): The maximum number of tokens to generate. Default is 128.
        """
        if self.endpoint.get("task") == EndpointTask.COMPLETIONS:  # completions endpoint
            if response_format:
                raise ValueError('response_format is not supported for completions endpoint')
            if system_prompt:
                prompt = f'{system_prompt} {prompt}'
            inputs = {
                'prompt': prompt,
                'temperature': temperature,
                'max_tokens': max_tokens,
            }
        else:  # chat endpoint
            sys_prompt_addition = f'\n\n{system_prompt}' if system_prompt else ''
            messages = [{
                'role': 'user',
                'content': prompt.replace('"', '').replace('“', '').replace('”', '') + sys_prompt_addition,
            }]
            if system_prompt:
                messages.insert(0, {
                    'role': 'system',
                    'content': system_prompt,
                })

            inputs = {
                'messages': messages,
                'temperature': temperature,
                'max_tokens': max_tokens,
            }

            if response_format:
                schema = response_format.model_json_schema()
                schema['additionalProperties'] = False
                inputs['response_format'] = {
                    'type': 'json_schema',
                    'json_schema': {
                        'name': 'structured_output',
                        'strict': True,
                        'schema': schema,
                    },
                }

        output = self.query(inputs)
        with self._token_lock:
            self.tokens_processed += output.total_tokens
        return output

    def generate_batch(
        self,
        prompts: List[str],
        system_prompt: Optional[str] = None,
        response_format: Optional[Type[BaseModel]] = None,
        temperature: float = 0.3,
        max_tokens: int = 1024,
        batch_size: int = 32,
        show_progress: bool = True,
        total_max_tokens: Optional[int] = None,
    ) -> List[GenerationOutput]:
        """
        Generate completions for a batch of prompts with optional token limit.
        Returns partial results if token limit is reached.
        """
        results: List[Tuple[int, GenerationOutput]] = []
        message_dict = dict(enumerate(prompts))
        pending_futures = set()

        @contextmanager
        def optional_progress(enabled: bool = True) -> Generator[Progress, Any, Any]:
            if enabled:
                with Progress(
                        TextColumn('[bold blue]{task.description}'),
                        BarColumn(bar_width=50),
                        TaskProgressColumn(),
                        TextColumn(
                            '•'),  # TODO: Likely want to remove the following because auto-refresh is off. We'll see
                        TimeElapsedColumn(),
                        TextColumn('•'),
                        TimeRemainingColumn(),
                        expand=False,
                        auto_refresh=False,
                        console=console) as progress:
                    yield progress
            else:

                class DummyConsole:
                    """Dummy console doesn't do anything"""

                    def print(self, *args, **kwargs):
                        print(*args, **kwargs)

                class DummyProgress:
                    """Dummy progress doesn't do anything"""
                    console = DummyConsole()

                    def add_task(self, *args, **kwargs):
                        del args, kwargs
                        return 0

                    def advance(self, *args, **kwargs):
                        del args, kwargs
                        pass

                    def update(self, *args, **kwargs):
                        del args, kwargs
                        pass

                    def stop(self):
                        pass

                yield cast(Progress, DummyProgress())

        single_query = partial(
            self.generate_completion,
            system_prompt=system_prompt,
            response_format=response_format,
            temperature=temperature,
            max_tokens=max_tokens,
        )

        total_tokens_processed = 0
        with optional_progress(show_progress) as progress:  # pylint: disable=contextmanager-generator-missing-cleanup
            task = progress.add_task('Generating outputs...', total=len(prompts))

            with ThreadPoolExecutor(max_workers=batch_size) as executor:
                future_to_index = {}
                for idx, message in message_dict.items():
                    future = executor.submit(single_query, message)
                    future_to_index[future] = idx
                    pending_futures.add(future)

                try:
                    for future in as_completed(pending_futures):
                        idx = future_to_index[future]
                        try:
                            response = future.result()
                            results.append((idx, response))
                            pending_futures.remove(future)
                            total_tokens_processed += response[1]

                            if total_max_tokens and total_tokens_processed > total_max_tokens:
                                progress.console.print('Token limit reached. Stopping further completions.')
                                break
                        except Exception as e:  # pylint: disable=broad-exception-caught
                            progress.console.print(f'Error processing message {idx}: {e}')
                            pending_futures.remove(future)
                            raise e
                        progress.advance(task)
                        progress.update(task, refresh=True)
                finally:
                    # Make sure we clean up any remaining futures
                    for future in pending_futures:
                        future.cancel()

                # Update progress to show actual completion
                progress.update(task, completed=len(results), refresh=True)

        return [r[1] for r in sorted(results, key=lambda x: x[0])]


class PayGoEndpoint(BaseEndpoint):
    """PayGo endpoint for model serving
    """

    def __init__(self, endpoint_name: str):
        self._endpoint_name = endpoint_name
        super().__init__()

    def setup_endpoint(self) -> DatabricksEndpoint:
        return self.client.get_endpoint(self._endpoint_name)


class ProvisionedThroughputEndpoint(BaseEndpoint):
    """
    Represents a provisioned throughput endpoint of a fine-tuned model that can generate text from a
    model. The endpoint is accessed through the Databricks Model Serving REST
    API and MLFlow Deployments library.

    Args:
        uc_model_path (str): The UC location of the model to be served. Ex: 'system.ai.dbrx_instruct'
        model_version (str): The version of the model to be served. Ex: '3'
        task_type (str): The task type of the model. Default is 'CHAT_COMPLETION'.
        scale_to_zero_enabled (bool): If True, the endpoint will scale to zero when not in use. Default is True.
        chunk_size_multiplier (int): The multiplier for the serving throughput chunk size. Default is 1.
        block_until_ready (bool): If True, the endpoint will block on returning until it is ready. Default is True.
    """

    def __init__(
        self,
        uc_model_path: str,
        model_version: Optional[str] = None,
        scale_to_zero_enabled: bool = True,
        chunk_size_multiplier: int = 1,
        block_until_ready: bool = True,
        teardown_on_exit: bool = True,
        warmup_after_ready_iters: int = 100,
    ) -> None:
        self._validate_inputs(uc_model_path, chunk_size_multiplier)

        if model_version is None:
            # Default to latest model version
            model_version = get_latest_model_version(uc_model_path)

        model_name = uc_model_path.split('.')[-1]
        self._endpoint_name = self._generate_endpoint_name(model_name, model_version)

        # PT Config
        throughput_chunk_size = self._get_throughput_chunk_size(uc_model_path, model_version)
        self.config = {
            'served_entities': [{
                'name': self._endpoint_name,
                'entity_name': uc_model_path,
                'entity_version': model_version,
                'workload_size': 'Small',
                'workload_type': 'GPU_SMALL',
                'scale_to_zero_enabled': scale_to_zero_enabled,
                'min_provisioned_throughput': 0,
                'max_provisioned_throughput': chunk_size_multiplier * throughput_chunk_size,
            }],
        }

        self.block_until_ready = block_until_ready
        self.teardown_on_exit = teardown_on_exit
        self.warmup_after_ready_iters = warmup_after_ready_iters

        super().__init__()

    def setup_endpoint(self) -> DatabricksEndpoint:
        self.endpoint = self.start_pt_endpoint()

        if not self.is_endpoint_ready():
            if self.block_until_ready:
                logger.info(f'Provisioned throughput endpoint is starting at {self.endpoint.name}. ' +
                            'Waiting for it to be ready...')
                self.wait_for_pt_endpoint_ready()
                logger.info(f'Provisioned throughput endpoint is ready at {self.endpoint.name}')
            else:
                print(f'Provisioned throughput endpoint is starting at {self.endpoint.name}. '
                      'Please wait for endpoint to be ready with '
                      f'{self.__class__.wait_for_pt_endpoint_ready.__name__} before use.')

        return self.endpoint

    def __exit__(self, _exc_type, _exc_value, _traceback) -> None:
        if self.teardown_on_exit:
            self.teardown_pt_endpoint()

    @staticmethod
    def _validate_inputs(uc_model_path: str, chunk_size_multiplier: int) -> None:
        """
        Validates input arguments for initialization.
        """
        model_path_regex = r'^[^/. ]+\.[^/. ]+\.[^/. ]+$'
        if not re.match(model_path_regex, uc_model_path):
            raise ValueError(
                f'Invalid uc_model_path {uc_model_path}. ' \
                "Please specify a UC location in the format '<catalog>.<schema>.<model>'. " \
                "This should be indicated under the 'Details' section on the model's page in your workspace.")

        if chunk_size_multiplier <= 0:
            raise ValueError('Please specify a chunk size multiplier greater than 0.')

    @staticmethod
    def _generate_endpoint_name(model_name: str, model_version: str) -> str:
        """
        Helper function that generates a unique endpoint name for the PT
        endpoint. Name is unique per-user and model to avoid conflicts.
        """
        workspace_id = str(get_me())
        hashed_name = md5_hash(workspace_id, num_digits=8)
        # rules: must be alphanumeric w/ hyphens and less than 64 characters
        endpoint_name = '-'.join([model_name[:44], f'v{model_version[:3]}', 'eval', hashed_name])
        endpoint_name = ''.join([c if c.isalnum() or c == '-' else '-' for c in endpoint_name])
        return endpoint_name

    @staticmethod
    def _get_throughput_chunk_size(uc_model_path: str, model_version: str) -> int:
        """
        Makes an request to the `serving_endpoints` API endpoint to get the
        optimized throughput chunk size. This is required for PT serving of
        fine-tuned models.
        """
        host, token = get_host_and_token_from_env()

        base_url = f'{host}/api/2.0/serving-endpoints'
        optimization_info_url = base_url + f'/get-model-optimization-info/{uc_model_path}/{model_version}'
        request_headers = {
            'Context-Type': 'text/json',
            'Authorization': f'Bearer {token}',
        }
        response = requests.get(url=optimization_info_url, headers=request_headers, timeout=10)
        if response.status_code != 200:
            raise HTTPError(f'Failed to retrieve throughput chunk size. Error {response.status_code}: {response.text}')

        throughput_information = response.json()
        throughput_chunk_size = throughput_information.get('throughput_chunk_size')

        # If `optimizable` is False, then we can't get `throughput_chunk_size`
        # because the model isn't supported by Databricks Model Serving.
        if throughput_chunk_size is None:
            optimizable = throughput_information.get('optimizable', False)
            if not optimizable:
                raise ValueError('Please use a model that is currently supported by Databricks Model serving: '
                                 'https://docs.databricks.com/en/machine-learning/foundation-models/index.html'
                                 '#provisioned-throughput-foundation-model-apis')
            raise ValueError('Could not retrieve throughput chunk size from '
                             '`serving-endpoints/model-optimization-info`. '
                             f'{json.dumps(throughput_information)}')

        return throughput_chunk_size

    def start_pt_endpoint(self) -> DatabricksEndpoint:
        """
        Spins up a provisioned throughput endpoint on behalf of the user's
        provided model.
        """
        try:

            existing_endpoint = self.client.get_endpoint(self._endpoint_name)
            print(f'Endpoint {self._endpoint_name} already exists. Skipping endpoint creation. '
                  f"Please wait for endpoint to be ready with '{self.__class__.wait_for_pt_endpoint_ready.__name__}' "
                  'before use.')
            return existing_endpoint
        except HTTPError as e:
            if e.response.status_code != 404:
                raise e

        # Create the endpoint through a POST request
        endpoint = self.client.create_endpoint(name=self._endpoint_name, config=self.config)

        return endpoint

    def wait_for_pt_endpoint_ready(self, timeout_mins: int = 30, check_interval_secs: int = 30) -> DatabricksEndpoint:
        """
        Waits for the PT endpoint to be ready before returning to the user.
        """
        t0 = time.time()
        timeout_secs = timeout_mins * 60
        while time.time() - t0 < timeout_secs:
            # refresh endpoint state
            try:
                self.endpoint = self.client.get_endpoint(self.endpoint.name)
            except HTTPError as e:
                if e.response.status_code == 404:
                    print(f'Endpoint {self.endpoint.name} does not exist. Please validate your served model endpoint '
                          'and spin it up manually if necessary.')
                    return self.endpoint
            state = self.endpoint.state
            if state['ready'] == 'READY' and state['config_update'] == 'NOT_UPDATING':
                if self.warmup_after_ready_iters > 0:
                    for _ in range(self.warmup_after_ready_iters):
                        try:
                            self.generate_completion('TESTING ENDPOINT READY', max_tokens=5)
                        except HTTPError as e:
                            if not e.response.status_code == 404:
                                raise e
                self.tokens_processed = 0
                return self.endpoint
            if state['config_update'] == 'UPDATE_FAILED':
                raise RuntimeError('Endpoint update failed. Please validate your served model endpoint and spin'
                                   ' it down manually if necessary.')
            if state['config_update'] == 'UPDATE_CANCELED':
                raise RuntimeError('Endpoint update was canceled. Please validate your served model endpoint and spin'
                                   ' it down manually if necessary.')
            time.sleep(check_interval_secs)

        raise TimeoutError(f'Provisioned throughput endpoint was not ready within {timeout_mins} minutes.')

    def teardown_pt_endpoint(self):
        """
        Tears down the provisioned throughput endpoint.
        """
        if self.is_endpoint_ready():
            # TODO: I think this should always delete the endpoint, even if it's not ready
            self.client.delete_endpoint(self.endpoint.name)
            logger.info(f'Endpoint {self.endpoint.name} has been torn down.')
        else:
            logger.warning(f'Endpoint {self.endpoint.name} is not ready. Skipping teardown.')

    def query(self,
              input_data: dict,
              max_retries: int = DEFAULT_MAX_RETRIES,
              backoff_exponential: float = DEFAULT_BACKOFF_EXPONENTIAL) -> GenerationOutput:
        """
        Queries the PT endpoint with the given input.

        Args:
            input_data (dict): Input data (or arguments) to pass to the deployment or model endpoint for inference.
        """

        if not self.is_endpoint_ready():
            raise RuntimeError('Provisioned throughput endpoint is not ready.'
                               'Please start it with `start_pt_endpoint`.')

        return super().query(input_data, max_retries, backoff_exponential)

    def generate_batch(
        self,
        *args,
        **kwargs,
    ) -> List[GenerationOutput]:

        if not self.is_endpoint_ready():
            raise RuntimeError('Provisioned throughput endpoint is not ready.'
                               'Please start it with `start_pt_endpoint`.')

        self._set_endpoint_class(EndpointClass.BATCH)
        responses = super().generate_batch(*args, **kwargs)
        self._set_endpoint_class(EndpointClass.QUERY)
        return responses



"""Module to encapsulate LLM invocation logic as custom MLFLow PythonModel."""

from typing import Dict, List, Optional

import mlflow
from mlflow.models import ModelSignature
from mlflow.pyfunc import PythonModelContext

from databricks.kie.inference_utils import get_llm_proxy_chat_completion_response
from databricks.kie.t2t_utils import (create_chat_completions_messages_from_instruction,
                                      create_chat_completions_request, generate_cot_response)


class CustomModel(mlflow.pyfunc.PythonModel):
    """
    A custom MLflow Python model that processes inputs based on provided instructions.

    Args:
        instruction (str): The instruction guiding the model's behavior.
        example_data (dict): A dictionary containing 'request' and 'response' examples for model signature inference.
        model_id (str): The identifier of the underlying language model to use for generating predictions.
    """

    def __init__(self,
                 instruction: str,
                 signature: ModelSignature,
                 model_id: str,
                 few_shot_examples: Optional[List[Dict[str, str]]] = None,
                 use_cot: bool = False,
                 model_descriptor: Optional[str] = None):
        self.instruction = instruction
        self.model_id = model_id
        self.few_shot_examples = few_shot_examples or []

        if model_descriptor is not None:
            self.model_descriptor = model_descriptor
        else:
            suffix = f"fewshot_{len(self.few_shot_examples)}" if len(self.few_shot_examples) > 0 else "zero_shot"
            self.model_descriptor = f"{model_id}_{suffix}"

        self.signature = signature
        self.use_cot = use_cot

    def predict(self, context: PythonModelContext, model_input: str, params: Optional[Dict] = None):
        del context, params
        messages = create_chat_completions_messages_from_instruction(self.instruction, model_input)
        if self.use_cot:
            return generate_cot_response(self.model_id, messages)

        req = create_chat_completions_request(messages)
        return get_llm_proxy_chat_completion_response(self.model_id, req)


class ModelWrapper():

    def __init__(self,
                 custom_model: CustomModel,
                 model_run_id: str,
                 run_name: str,
                 registered_model_name: Optional[str] = None):
        self.custom_model = custom_model
        self.run_id = model_run_id
        self.run_name = run_name
        self.registered_model_name = registered_model_name

    def get_model_path(self):
        return f"runs:/{self.run_id}/model"



"""Module that encapsulates instruction analysis for Text2Text."""

from typing import Dict, List

from databricks.kie.inference_utils import generate_base_model_using_chat_completion_messages
from databricks.kie.t2t_schema import InstructionInfo, InstructionType

INSTRUCTION_INFO_GENERATION_WITHOUT_ICL = """
You are an expert in instruction analysis and generation. That is, given 
- the original instruction provided by the user
- List of criteria that response should satisfy
you should `infer` the OPTIMAL instruction to generate the response given the request. Instruction should be as detailed as possible.

The instruction info to generate has the following fields:
- optimized_instruction: string instruction that can be used to create the response given the request.
- output_format: Format of the output string. This should detail formatting constraints of the output. 
        Example output formats includes guidlines such as:
        e.g)    - 'Output should be a comma separated list'
                - 'Response should be as concise as possible'
                - 'Response should be json with keys a, b, c'
                - 'Response should not contains phrases like...'
                - 'Response should have the tone of ...'

- instruction_type: Enum that describes the type of instruction. Can be one of : {instruction_type_list}

The original instruction provided by the user is provided after the header `original_instruction`.
The list of criteria response must satisfy is provided after the header `response_criteria`

original_instruction: {instruction}
response_criteria: {response_criteria}
"""

INSTRUCTION_INFO_GENERATION_WITH_ICL = """
You are an expert in instruction analysis / generation. That is, given 
  - a list of JSON objects containing `request` and `response` keys
  - the original instruction provided by the user
  - List of criteria that response should satisfy

you should `infer` the OPTIMAL instruction to generate the response given the request. Instruction should be as detailed as possible.
The instruction info to generate has following fields.

- optimized_instruction: string instruction that can be used to create the response given the request.
- output_format: Format of the output string. This should detail formatting constraints of the output. 
        Example output formats includes guidlines such as:
        e.g)    - 'Output should be a comma separated list'
                - 'Response should be as concise as possible'
                - 'Response should be json with keys a, b, c'
                - 'Response should not contains phrases like...'
                - 'Response should have the tone of ...'

- instruction_type: Enum that describes the type of instruction. Can be one of : {instruction_type_list}

The original instruction provided by the user is provided after the header `original_instruction`.
The list of JSON objects containing `request` and `response` keys is provided after the header `example_data`.
The list of criteria response must satisfy is provided after the header `response_criteria`

example_data: {data}
original_instruction: {instruction}
response_criteria: {response_criteria}
"""


def generate_instruction_info(example_data: List[Dict[str, str]], instruction: str,
                              eval_criteria: List[str]) -> InstructionInfo:
    instruction_type_list = f"{ ', '.join(str(v.name) for v in InstructionType) }"

    if example_data:
        system_msg = INSTRUCTION_INFO_GENERATION_WITH_ICL.format(data=example_data,
                                                                 instruction=instruction,
                                                                 instruction_type_list=instruction_type_list,
                                                                 response_criteria=eval_criteria)
    else:
        system_msg = INSTRUCTION_INFO_GENERATION_WITHOUT_ICL.format(instruction=instruction,
                                                                    instruction_type_list=instruction_type_list,
                                                                    response_criteria=eval_criteria)

    messages = [{"role": "system", "content": system_msg}]

    return generate_base_model_using_chat_completion_messages(messages,
                                                              InstructionInfo,
                                                              model_id="gpt-4o-2024-08-06-text2text")



"""Module for creating Text2Text system."""
from typing import Dict, List, Optional

from databricks.kie.t2t_runner import Text2TextRunner
from databricks.kie.t2t_schema import JsonlFileDataConfig, Text2TextTaskSpec, UCTableDataConfig


class Text2TextFactory:
    """Factory class for creating Text2Text system."""

    @staticmethod
    def from_jsonl(instruction: str,
                   file_path: str,
                   input_key_name: str,
                   output_key_name: str,
                   experiment_name: Optional[str] = None,
                   evaluation_criteria: Optional[List[str]] = None):
        """Creates a Text2Text system from a jsonl file in UC Volume."""
        task_spec = Text2TextTaskSpec(instruction=instruction,
                                      jsonl_file=JsonlFileDataConfig(file_path=file_path,
                                                                     input_key=input_key_name,
                                                                     output_key=output_key_name),
                                      experiment_name=experiment_name,
                                      evaluation_criteria=evaluation_criteria)

        return Text2TextRunner(task_spec=task_spec)

    @staticmethod
    def from_uc_table(instruction: str,
                      table_path: str,
                      input_column_name: str,
                      output_column_name: str,
                      experiment_name: Optional[str] = None,
                      evaluation_criteria: Optional[List[str]] = None):
        """Creates a Text2Text system from Unity Catalog table."""
        task_spec = Text2TextTaskSpec(instruction=instruction,
                                      uc_table=UCTableDataConfig(table_path=table_path,
                                                                 input_column_name=input_column_name,
                                                                 output_column_name=output_column_name),
                                      experiment_name=experiment_name,
                                      evaluation_criteria=evaluation_criteria)
        return Text2TextRunner(task_spec=task_spec)

    @staticmethod
    def from_examples(instruction: str,
                      examples: List[Dict[str, str]],
                      experiment_name: Optional[str] = None,
                      evaluation_criteria: Optional[List[str]] = None):
        """Creates a Text2Text system from a list of examples."""
        task_spec = Text2TextTaskSpec(instruction=instruction,
                                      json_examples=examples,
                                      experiment_name=experiment_name,
                                      evaluation_criteria=evaluation_criteria)
        return Text2TextRunner(task_spec=task_spec)



"""Utility functions for abstract out request construction / response parsing of Text-to-Text tasks."""

import json
from typing import Dict, List, Optional

from databricks.sdk import WorkspaceClient  # pylint: disable = ungrouped-imports
from databricks.sdk.errors import ResourceDoesNotExist  # pylint: disable = ungrouped-imports
from openai.lib._parsing._completions import type_to_response_format_param
from pydantic import BaseModel, Field

from databricks.kie.inference_utils import (generate_base_model_using_chat_completion_messages,
                                            get_llm_proxy_chat_completion_response)

INSTRUCTION_AND_INPUT_PROMPT_TEMPLATE = """## User input : {inp}

## Instruction : {instruction}
"""

LLM_JUDGE_EVAL_PROMPT_TEMPLATE = """
Your task is to evaluate and determine whether the provided `response` meets the below `evaluation criteria`. 
You are given input instruction, response, and evaluation criteria.
input instruction is provided following the header: `input`.
response is provided following the header `response`.
evaluation criteria is provided following the header `evaluation criteria`.

You should output a 5 if there `response` perfectly met/satisfied the evaluation critieria, and 1 if response did not meet the evaluation criteria.

evaluation criteria: '{eval_criteria}'
"""

EVAL_CRITERIA_GENERATOR_SYSTEM_PROMPT = (
    "You are an expert at user instruction analysis. "
    "Given an instruction for any task provided by the user, return a JSON list of criteria."
    "The criteria will be used to score an LLM on how well it followed the instructions given a specific input."
    "Provide detailed granular evaluation criteria on fomatting, style, and correctness."
    "Instruction to analyze is provided in in user message following 'Instruction' header."
    "Provide detailed granular evaluation criteria on fomatting, style, and correctness.")

EVAL_CRITERIA_GENERATOR_SYSTEM_PROMPT_WITH_EXAMPLES = (
    "You are an expert at user instruction analysis. "
    "Given an instruction for any task provided by the user and a JSON containing ground truth request"
    " and response pairs, return a JSON list of criteria."
    "The criteria will be used to score an LLM on how well it followed the instructions given a specific input."
    "Instruction to analyze is provided in in user message following 'Instruction' header, "
    "and examples are provided in user message following 'Ground Truth Examples' header."
    "Provide detailed granular evaluation criteria on fomatting, style, and correctness.")

SYSTEM_PROMPT_TEMPLATE = """
You are a helpful AI Agent. Please follow the following instructions with given input.
The instruction provided by the user is included following the "Instruction" header below.
The instruction describes the general task that the user would like to accomplish.

Instruction : '{instruction}'

The specific input for this request is included in the as part of user message.
"""

USER_PROMPT_TEMPLATE = """
user input : '{inp}'
"""

EVAL_PROMPT_TEMPLATE = """
Your task is to evaluate the response using the following criteria given input and response.

evaluation criteria: '{eval_criteria}'
"""

EVAL_PROMPT_SUFFIX = """
 This is the input: '{request}'
 This is the model output to evaluate: '{response}'
"""

CHAIN_OF_THOUGHT_SYSTEM_PROMPT = """
Given instruction and user input provided as the system message and user messages above, generate CoT class.

CoT class provide a detailed step-by-step reasoning that leads to answering the provided user instruction.
CoT class has 2 fields: `step_by_step_reasoning` and `concise_response`.

step_by_step_reasoning : please provide detailed step-by-step reasoning that leads to answering the provided user instruction.
concise_response: please provide the most correct and concise answer to the user instruction.

If few shot examples are provided in the user messages above, refer to assistant messages in few shot examples
to infer the `concise_response` fields. Note that previous assistant output to few shot examples does not
include the `step_by_step_reasoning`. If few shot examples exists as messages above, refer to them as
examples for generating the `concise_response` field.
"""


class EvaluationCriteria(BaseModel):
    criteria: List[str] = Field(description="List of evaluation criteria that the response must satisfy.")


class CoT(BaseModel):
    step_by_step_reasoning: str = Field(
        description="Detailed step by step reasoning that leads to answering the provided user instruction.")
    concise_response: str = Field(description="The most correct and concise answer to the user instruction.")


def create_chat_completions_request(messages: List[Dict[str, str]], response_format: Optional[Dict] = None):
    """
    Creates a chat completions request json.

    Args:
        messages (list): A list of message dictionaries to be included in the request.
        response_format (str, optional): The desired format of the response. Defaults to None.

    Returns:
        dict: A dictionary representing the chat completions request.
    """
    chat_completion_req: Dict = {
        "messages": messages,
    }
    if response_format:
        chat_completion_req['response_format'] = response_format
    return chat_completion_req


def create_chat_completions_messages_from_instruction(instruction: str, inp: str):
    """
    Creates a list of chat completion messages based on the given instruction and input.

    Args:
        instruction (str): The instruction to be included in the system message.
        inp (str): The input to be included in the user message.

    Returns:
        list: A list of dictionaries representing the chat messages. Each dictionary contains
              a 'role' key (either 'system' or 'user') and a 'content' key with the respective message.
    """

    messages = [{
        "role": "system",
        "content": SYSTEM_PROMPT_TEMPLATE.format(instruction=instruction)
    }, {
        "role": "user",
        "content": USER_PROMPT_TEMPLATE.format(inp=inp)
    }]
    return messages


def generate_evaluation_criteria(instruction: str, examples: Optional[List[Dict]] = None) -> List[str]:
    """Generates list of evaluation criteria based on the given instruction and examples.

    Args:
        instruction (str): The instruction to generate evaluation criteria for.
        examples (list, optional): A list of dictionaries containing ground truth request and response pairs.
            Defaults to None.

    Returns:
        list: A list of EvaluationCriterion objects.
    """
    messages = [{
        "role":
            "system",
        "content":
            EVAL_CRITERIA_GENERATOR_SYSTEM_PROMPT_WITH_EXAMPLES if examples else EVAL_CRITERIA_GENERATOR_SYSTEM_PROMPT
    }, {
        "role":
            "user",
        "content": (f"Instruction: {instruction}\n\n Ground Truth Examples: {json.dumps(examples)}"
                    if examples else f"Instruction: {instruction}")
    }]

    return generate_base_model_using_chat_completion_messages(messages,
                                                              EvaluationCriteria,
                                                              model_id="gpt-4o-2024-08-06-text2text").criteria


def generate_cot_response(model_id: str, messages: List[Dict]) -> str:
    """
    Generates a response that has gone through CoT reasoning.
    
    Note: CHAIN of thought reasoning is not returned -- it is used only for improving reasoning
    as model autoregressively generates the response tokens.
    The implementation is structured such that chain of thought reasoning is always generated first.

    Args:
        model_id (str): The model identifier to use for generating the response.
        messages (list): A list of message dictionaries to be included in the request.
            Corresponds to chat completion messges.
    
    Returns:
        str: The most correct and concise answer to the user instruction
    """
    messages.append({"role": "system", "content": CHAIN_OF_THOUGHT_SYSTEM_PROMPT})
    req = create_chat_completions_request(messages, type_to_response_format_param(CoT))
    res = get_llm_proxy_chat_completion_response(model_id, req)
    cot_response = CoT(**json.loads(res))
    return cot_response.concise_response


def validate_secret_exists(secret_scope: str, secret_key: str) -> bool:
    """
    Validates if the secret exists in the secret scope.

    Args:
        secret_scope (str): The secret scope name.
        secret_key (str): The secret key name.

    Returns:
        bool: True if the secret exists, False otherwise.
    """
    w = WorkspaceClient()
    try:
        w.secrets.get_secret(scope=secret_scope, key=secret_key)
        return True
    except ResourceDoesNotExist:
        return False



"""Get model training runs"""

from typing import Union

from databricks.model_training.api.exceptions import DatabricksGenAIResponseError
from databricks.model_training.api.foundation_model.list import list as get_training_runs_paginated
from databricks.model_training.types import TrainingRun


def get(training_run: Union[str, TrainingRun]) -> TrainingRun:
    """Get a single training run by name or run object

    Args:
        training_run (Union[str, TrainingRun]): The training run to get.

    Returns:
        TrainingRun: The training run
    """
    training_runs = [training_run] if isinstance(training_run, str) else [training_run.name]
    run = get_training_runs_paginated(
        training_runs=training_runs,
        include_details=True,
    )

    if not run:
        name = training_run if isinstance(training_run, str) else training_run.name
        raise DatabricksGenAIResponseError(f'Finetuning run {name} not found')

    return run[0]



"""Logging utilities"""
import logging
from contextlib import contextmanager

from rich.console import Console

from databricks.model_training.api.utils import is_running_in_databricks_notebook


def get_console() -> Console:
    if is_running_in_databricks_notebook():
        return Console(force_jupyter=True)
    return Console()


console = get_console()


@contextmanager
def temp_log_level(name: str, level: int):
    logger = logging.getLogger(name)
    # Save old level
    old_level = logger.getEffectiveLevel()
    # Set new level
    logger.setLevel(level)
    try:
        yield
    finally:
        # Restore old level
        logger.setLevel(old_level)



"""
Training types
"""
from .training_constants import TrainingConstants, TrainingModelConstants
from .training_run import TrainingEvent, TrainingRun

__all__ = [
    'TrainingModelConstants',
    'TrainingConstants',
    'TrainingRun',
    'TrainingEvent',
]



"""
Databricks training constants
"""

from dataclasses import dataclass
from http import HTTPStatus
from typing import Any, Dict, List, Tuple

from mcli.api.schema.generic_model import DeserializableModel
from mcli.utils.utils_string_functions import camel_case_to_snake_case

from databricks.model_training.api.engine.engine import MAPIException


@dataclass()
class TrainingModelConstants(DeserializableModel):
    """Databricks model constants

    Args:
        name: The name of the model
        display_name: A more `displayable` name for the model
        max_context_length: Maximum context length of the model (tokens)
    """

    name: str
    display_name: str
    max_context_length: int

    _required_properties: Tuple[str, ...] = tuple([
        'name',
        'displayName',
        'maxContextLength',
    ])

    @classmethod
    def from_mapi_response(cls, response: Dict[str, Any]) -> 'TrainingModelConstants':
        """Load the model info from MAPI response.
        """
        missing = set(cls._required_properties) - set(response)

        if missing:
            raise MAPIException(
                status=HTTPStatus.BAD_REQUEST,
                message='Missing required key(s) in response to deserialize TrainingModelConstants'
                f'object: {", ".join(missing)} ',
            )

        args = {camel_case_to_snake_case(prop): response.get(prop) for prop in cls._required_properties}
        return cls(**args)

    def _repr_html_(self) -> str:
        """Display the model card as HTML.
        """
        keys = ['name', 'display_name', 'max_context_length']

        attr_to_label = {key: ' '.join([word.title() for word in key.split('_')]) for key in keys}

        # pylint: disable-next=import-outside-toplevel,cyclic-import
        from databricks.model_training.types.common import generate_vertical_html_table
        return generate_vertical_html_table([self], attr_to_label)


@dataclass
class TrainingConstants(DeserializableModel):
    """Contains a list of Databricks model constants for training

    Args:
        models: A list of model cards
    """
    models: List[TrainingModelConstants]

    _required_properties: Tuple[str, ...] = tuple([
        'models',
    ])

    @classmethod
    def from_mapi_response(cls, response: Dict[str, Any]) -> 'TrainingConstants':
        """Load the model info list from MAPI response.
        """

        missing = set(cls._required_properties) - set(response)

        if missing:
            raise MAPIException(
                status=HTTPStatus.BAD_REQUEST,
                message='Missing required key(s) in response to deserialize TrainingConstants'
                f'object: {", ".join(missing)} ',
            )

        return cls(**{
            'models': [TrainingModelConstants.from_mapi_response(model_info) for model_info in response.get('models')]
        })



"""
For training foundation models
"""
from .cancel import cancel
from .create import create
from .get import get
from .get_checkpoints import get_checkpoints
from .get_events import get_events
from .get_models import get_models
from .list import list  # pylint: disable=redefined-builtin

__all__ = [
    'get_checkpoints',
    'get_models',
    'cancel',
    'create',
    'get',
    'list',
    'get_events',
]



"""Prepare the user to run ai_query on their data"""

import os
from typing import Tuple, cast

import mlflow
from mlflow import MlflowException
from mlflow.entities import Run as MlflowRun

from databricks.kie.constants import MLFLOW_REGISTERED_MODEL_TAG, MLFLOW_SYSTEM_PROMPT_FILE, MLFLOW_TRAILING_PROMPT_FILE
from databricks.kie.data_utils import get_all_unlabeled
from databricks.kie.kie_state import KIEState
from databricks.kie.task_spec import KIETaskSpec
from databricks.model_serving.types.pt_endpoint import ProvisionedThroughputEndpoint
from databricks.model_training.api.utils import check_if_table_exists, check_table_has_columns


def prepare_output_table(task_spec: KIETaskSpec, state: KIEState):
    """Prepare the output table for ai_query"""

    if check_if_table_exists(task_spec.output_table) and check_table_has_columns(task_spec.output_table, ("request",)):
        print(f"Output table {task_spec.output_table} already exists with the correct schema.")
        return

    # Read the data into a DataFrame and write it to a Delta table
    print("Preparing output table for ai_query")
    print(f"Reading documents into {task_spec.output_table}. This may take a moment...")
    df = get_all_unlabeled(state.unlabeled_split_df)
    df.write.mode("overwrite").saveAsTable(task_spec.output_table)


def get_model_from_run(run: MlflowRun) -> str:
    """Get the model UC schema from an mlflow run based on the registered_to tag
    """
    # Get the model from the run
    if MLFLOW_REGISTERED_MODEL_TAG not in run.data.tags:
        raise ValueError("This run is not registered to a model through the KIE notebook."
                         f"Set a tag on the run with `{MLFLOW_REGISTERED_MODEL_TAG}` to the UC schema of the model.")

    uc_schema = run.data.tags[MLFLOW_REGISTERED_MODEL_TAG]
    return uc_schema


def get_prompts_from_run(run: MlflowRun) -> Tuple[str, str]:
    """Get the system and trailing prompts from an mlflow run
    """
    if not run.info.artifact_uri:
        raise ValueError("This run does not have an artifact URI")

    try:
        root_uri = cast(str, run.info.artifact_uri)
        system_prompt = mlflow.artifacts.load_text(os.path.join(root_uri, MLFLOW_SYSTEM_PROMPT_FILE))
        trailing_prompt = mlflow.artifacts.load_text(os.path.join(root_uri, MLFLOW_TRAILING_PROMPT_FILE))
    except MlflowException as e:
        raise ValueError(
            "This run does not have the system and trailing prompts saved as artifacts."
            f"Make sure the prompts are saved as `{MLFLOW_SYSTEM_PROMPT_FILE}` and `{MLFLOW_TRAILING_PROMPT_FILE}`"
        ) from e

    return system_prompt, trailing_prompt


def ready_pt_endpoint_from_run(run: MlflowRun) -> ProvisionedThroughputEndpoint:
    """Create a PT endpoint from an mlflow run
    """
    # Get the tag on the mlflow run
    model_uc_schema = get_model_from_run(run)

    # Create the PT endpoint
    endpoint = ProvisionedThroughputEndpoint(model_uc_schema, block_until_ready=False)

    # Create the PT endpoint
    return endpoint


def ready_pt_endpoint_from_model(model_uc_schema: str) -> ProvisionedThroughputEndpoint:
    """Create a PT endpoint from a model UC schema
    """
    return ProvisionedThroughputEndpoint(model_uc_schema, block_until_ready=False)



"""List events for a model training run"""

from time import sleep
from typing import Union, cast

from IPython import get_ipython  # type: ignore
from IPython.core.display import HTML, clear_output, display

from databricks.model_training.api.engine import get_return_response, run_plural_mapi_request
from databricks.model_training.types import TrainingEvent, TrainingRun
from databricks.model_training.types.common import ObjectList

QUERY_FUNCTION = 'getFinetuneEvents'
VARIABLE_DATA_NAME = 'getFinetuneEventsData'
# This returns the same data that the create_run function returns
# for consistency when rendering the describe output

QUERY = f"""
query GetFinetuneEvents(${VARIABLE_DATA_NAME}: GetFinetuneEventsInput!) {{
  {QUERY_FUNCTION}({VARIABLE_DATA_NAME}: ${VARIABLE_DATA_NAME}) {{
    eventType
    eventTime
    eventMessage
  }}
}}"""


def is_running_in_notebook() -> bool:
    """
    Tested that this holds true for Databricks, Colab, and Jupyter notebooks
    Tested this is false when ran from python or from an ipython shell.

    Returns:
        bool: True if running in a notebook, False otherwise.
    """
    try:
        # Check if not in IPython shell
        if 'IPKernelApp' not in get_ipython().config:  # type: ignore
            return False
    except Exception:  # pylint: disable=W0718
        return False
    return True


def get_events(training_run: Union[str, TrainingRun], follow: bool = False) -> ObjectList[TrainingEvent]:
    """List model training runs

    Args:
        training_run (Union[str, TrainingRun]): The training run to get events for.
        follow (bool): Follow the events for the run while it is in-progress.

    Returns:
        List[TrainingEvent]: A list of training run events. Each event has an event
            type, time, and message.
    """
    training_run_name = training_run.name if isinstance(training_run, TrainingRun) else training_run

    variables = {
        VARIABLE_DATA_NAME: {
            'name': training_run_name,
        },
    }

    if follow:
        ft_events = cast(ObjectList[TrainingEvent], [])
        while len(ft_events) == 0 or not ft_events[-1].is_terminal_state():  # pylint: disable=unsubscriptable-object,no-member
            new_events = get_events(training_run)
            if len(new_events) > len(ft_events):
                if is_running_in_notebook():
                    clear_output()
                    display(HTML(new_events._repr_html_()))  # pylint: disable=protected-access
                else:
                    for event in new_events[len(ft_events):]:
                        print(event)

            ft_events = new_events

            # don't unnecessarily sleep if run has completed
            if not ft_events[-1].is_terminal_state():
                sleep(5)

        # final cleanup before displaying in notebook
        if is_running_in_notebook():
            clear_output()

    else:
        response = run_plural_mapi_request(
            query=QUERY,
            query_function=QUERY_FUNCTION,
            return_model_type=TrainingEvent,
            variables=variables,
        )
        ft_events = get_return_response(response)

    return ft_events
